/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "g_local.h"
#include "m_player.h"



static	edict_t		*current_player;
static	gclient_t	*current_client;

trace_t pmtrace;

static	vec3_t	forward, right, up;
float	xyspeed;

float	bobmove;
int		bobcycle;		// odd cycles are right foot going forward
float	bobfracsin;		// sin(bobfrac*M_PI)

/*
===============
SV_CalcRoll

===============
*/
float SV_CalcRoll (vec3_t velocity)
{
	float	sign;
	float	side;
	float	value;

	side = DotProduct (velocity, right);
	sign = side < 0 ? -1 : 1;
	side = fabs(side);

	value = sv_rollangle->value;

	if (side < sv_rollspeed->value)
		side = side * value / sv_rollspeed->value;
	else
		side = value;

	return side*sign;

}


/*
===============
P_DamageFeedback

Handles color blends and view kicks
===============
*/
void P_DamageFeedback (edict_t *player)
{
	gclient_t	*client;
	float	side;
	float	realcount, count, kick;
	vec3_t	v;
	static	vec3_t	power_color = {0.0, 1.0, 0.0};
	static	vec3_t	acolor = {1.0, 1.0, 1.0};
	static	vec3_t	bcolor = {1.0, 0.0, 0.0};

	client = player->client;

	// flash the backgrounds behind the status numbers
	client->ps.stats[STAT_FLASHES] = 0;
	if (client->damage_blood)
		client->ps.stats[STAT_FLASHES] |= 1;
	if (client->damage_armor && !(player->flags & FL_GODMODE) && (client->invincible_framenum <= level.framenum))
		client->ps.stats[STAT_FLASHES] |= 2;

	// total points of damage shot at the player this frame
	count = (client->damage_blood + client->damage_armor + client->damage_parmor);
	if (count == 0)
		return;		// didn't take any damage

	// start a pain animation if still in the player model

	//splitmodels (PAIN) jal [start]
	if (player->client->anim_priority < ANIM_PAIN && player->s.modelindex == 255)
	{
		static int		i;

		player->client->anim_priority = ANIM_PAIN;
		// removed crouched variants

		i = (i+1)%3;
		switch (i)
		{
		case 0:
			player->s.frame = (TORSO_PAIN1 & 0x3F);
			player->client->anim_event_time = TORSO_PAIN1_time;

			break;
		case 1:
			player->s.frame = (TORSO_PAIN2 & 0x3F);
			player->client->anim_event_time = TORSO_PAIN2_time;

			break;
		case 2:
			player->s.frame = (TORSO_PAIN1 & 0x3F);
			player->client->anim_event_time = TORSO_PAIN1_time;

			break;
		}
	}
	//jal [end]

	realcount = count;
	if (count < 10)
		count = 10;	// always make a visible effect

	// play an apropriate pain sound
	if(!sv_godall->value)
	{
		if ((level.time > player->pain_debounce_time) && !(player->flags & FL_GODMODE) && (client->invincible_framenum <= level.framenum))
		{
			player->pain_debounce_time = level.time + 0.7;

			if (player->health < 25)
				player->s.event = EV_PAIN25;
			else if (player->health < 50)
				player->s.event = EV_PAIN50;
			else if (player->health < 75)
				player->s.event = EV_PAIN75;
			else
				player->s.event = EV_PAIN100;
		}
	}

	// the total alpha of the blend is always proportional to count
	if (client->damage_alpha < 0)
		client->damage_alpha = 0;
	client->damage_alpha += count*0.01;
	if (client->damage_alpha < 0.2)
		client->damage_alpha = 0.2;
	if (client->damage_alpha > 0.6)
		client->damage_alpha = 0.6;		// don't go too saturated

	// the color of the blend will vary based on how much was absorbed
	// by different armors
	VectorClear (v);
	if (client->damage_parmor)
		VectorMA (v, (float)client->damage_parmor/realcount, power_color, v);
	if (client->damage_armor)
		VectorMA (v, (float)client->damage_armor/realcount,  acolor, v);
	if (client->damage_blood)
		VectorMA (v, (float)client->damage_blood/realcount,  bcolor, v);
	VectorCopy (v, client->damage_blend);


	//
	// calculate view angle kicks
	//
	kick = abs(client->damage_knockback);
	if (kick && player->health > 0)	// kick of 0 means no view adjust at all
	{
		kick = kick * 100 / player->health;

		if (kick < count*0.5)
			kick = count*0.5;
		if (kick > 50)
			kick = 50;

		VectorSubtract (client->damage_from, player->s.origin, v);
		VectorNormalize (v);

		side = DotProduct (v, right);
		client->v_dmg_roll = kick*side*0.3;

		side = -DotProduct (v, forward);
		client->v_dmg_pitch = kick*side*0.3;

		client->v_dmg_time = level.time + DAMAGE_TIME;
	}

	//
	// clear totals
	//
	client->damage_blood = 0;
	client->damage_armor = 0;
	client->damage_parmor = 0;
	client->damage_knockback = 0;
}




/*
===============
SV_CalcViewOffset

Auto pitching on slopes?

  fall from 128: 400 = 160000
  fall from 256: 580 = 336400
  fall from 384: 720 = 518400
  fall from 512: 800 = 640000
  fall from 640: 960 =

  damage = deltavelocity*deltavelocity  * 0.0001

===============
*/
void SV_CalcViewOffset (edict_t *ent)
{
	float		*angles;
	float		bob;
	float		ratio;
	float		delta;
	vec3_t		v;


//===================================

	// base angles
	angles = ent->client->ps.kick_angles;

	// if dead, fix the angle and don't add any kick
	if (ent->deadflag)
	{
		VectorClear (angles);

		ent->client->ps.viewangles[ROLL] = 40;
		ent->client->ps.viewangles[PITCH] = -15;
		ent->client->ps.viewangles[YAW] = ent->client->killer_yaw;
	}
	else
	{
		// add angles based on weapon kick

		VectorCopy (ent->client->kick_angles, angles);

		// add angles based on damage kick

		ratio = (ent->client->v_dmg_time - level.time) / DAMAGE_TIME;
		if (ratio < 0)
		{
			ratio = 0;
			ent->client->v_dmg_pitch = 0;
			ent->client->v_dmg_roll = 0;
		}
		angles[PITCH] += ratio * ent->client->v_dmg_pitch;
		angles[ROLL] += ratio * ent->client->v_dmg_roll;

		// add pitch based on fall kick

		ratio = (ent->client->fall_time - level.time) / FALL_TIME;
		if (ratio < 0)
			ratio = 0;
		angles[PITCH] += ratio * ent->client->fall_value;

		// add angles based on velocity

		delta = DotProduct (ent->velocity, forward);
		angles[PITCH] += delta*run_pitch->value;

		delta = DotProduct (ent->velocity, right);
		angles[ROLL] += delta*run_roll->value;

		// add angles based on bob

		delta = bobfracsin * bob_pitch->value * xyspeed;
		if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
			delta *= 6;		// crouching
		angles[PITCH] += delta;
		delta = bobfracsin * bob_roll->value * xyspeed;
		if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
			delta *= 6;		// crouching
		if (bobcycle & 1)
			delta = -delta;
		angles[ROLL] += delta;
	}

//===================================

	// base origin

	VectorClear (v);

	// add view height

	v[2] += ent->viewheight;

	// add fall height

	ratio = (ent->client->fall_time - level.time) / FALL_TIME;
	if (ratio < 0)
		ratio = 0;
	v[2] -= ratio * ent->client->fall_value * 0.4;

	// add bob height

	bob = bobfracsin * xyspeed * bob_up->value;
	if (bob > 6)
		bob = 6;
	//gi.DebugGraph (bob *2, 255);
	v[2] += bob;

	// add kick offset

	VectorAdd (v, ent->client->kick_origin, v);

	// absolutely bound offsets
	// so the view can never be outside the player box
	clamp ( v[0], -14, 14 );
	clamp ( v[1], -14, 14 );
	clamp ( v[2], -22, 30 );

	VectorCopy (v, ent->client->ps.viewoffset);
}

/*
==============
SV_CalcGunOffset
==============
*/
void SV_CalcGunOffset (edict_t *ent)
{
	int		i;
	float	delta;

	// gun angles from bobbing
	ent->client->ps.gunangles[ROLL] = xyspeed * bobfracsin * 0.005;
	ent->client->ps.gunangles[YAW] = xyspeed * bobfracsin * 0.01;
	if (bobcycle & 1)
	{
		ent->client->ps.gunangles[ROLL] = -ent->client->ps.gunangles[ROLL];
		ent->client->ps.gunangles[YAW] = -ent->client->ps.gunangles[YAW];
	}

	ent->client->ps.gunangles[PITCH] = xyspeed * bobfracsin * 0.005;

	// gun angles from delta movement
	for (i=0 ; i<3 ; i++)
	{
		delta = ent->client->oldviewangles[i] - ent->client->ps.viewangles[i];
		if (delta > 180)
			delta -= 360;
		if (delta < -180)
			delta += 360;
		if (delta > 45)
			delta = 45;
		if (delta < -45)
			delta = -45;
		if (i == YAW)
			ent->client->ps.gunangles[ROLL] += 0.1*delta;
		ent->client->ps.gunangles[i] += 0.2 * delta;
	}
}


/*
=============
SV_AddBlend
=============
*/
void SV_AddBlend (float r, float g, float b, float a, float *v_blend)
{
	float	a2, a3;

	if (a <= 0)
		return;
	a2 = v_blend[3] + (1-v_blend[3])*a;	// new total alpha
	a3 = v_blend[3]/a2;		// fraction of color from old

	v_blend[0] = v_blend[0]*a3 + r*(1-a3);
	v_blend[1] = v_blend[1]*a3 + g*(1-a3);
	v_blend[2] = v_blend[2]*a3 + b*(1-a3);
	v_blend[3] = a2;
}


/*
=============
SV_CalcBlend
=============
*/
void SV_CalcBlend (edict_t *ent)
{
	int		contents;
	vec3_t	vieworg;
	int		remaining;

	ent->client->ps.blend[0] = ent->client->ps.blend[1] =
		ent->client->ps.blend[2] = ent->client->ps.blend[3] = 0;

	// add for contents
	VectorAdd (ent->s.origin, ent->client->ps.viewoffset, vieworg);
	contents = gi.pointcontents (vieworg);

	if (contents & CONTENTS_LAVA)
		SV_AddBlend (1.0, 0.3, 0.0, 0.6, ent->client->ps.blend);
	else if (contents & CONTENTS_SLIME)
		SV_AddBlend (0.0, 0.1, 0.05, 0.6, ent->client->ps.blend);

	// add for powerups
	if (ent->client->quad_framenum > level.framenum)
	{
		remaining = ent->client->quad_framenum - level.framenum;
		if (remaining == 30)	// beginning to fade
			gi.sound(ent, CHAN_ITEM, gi.soundindex("sound/items/damage2.wav"), 1, ATTN_NORM, 0);
		if (remaining > 30 || (remaining & 4) )
			SV_AddBlend (0, 0, 1, 0.08, ent->client->ps.blend);
	}
	else if (ent->client->invincible_framenum > level.framenum)
	{
		remaining = ent->client->invincible_framenum - level.framenum;
		if (remaining == 30)	// beginning to fade
			gi.sound(ent, CHAN_ITEM, gi.soundindex("sound/items/protect2.wav"), 1, ATTN_NORM, 0);
		if (remaining > 30 || (remaining & 4) )
			SV_AddBlend (1, 1, 0, 0.08, ent->client->ps.blend);
	}
	else if (ent->client->enviro_framenum > level.framenum)
	{
		remaining = ent->client->enviro_framenum - level.framenum;
		if (remaining == 30)	// beginning to fade
			gi.sound(ent, CHAN_ITEM, gi.soundindex("sound/items/airout.wav"), 1, ATTN_NORM, 0);
		if (remaining > 30 || (remaining & 4) )
			SV_AddBlend (0, 1, 0, 0.08, ent->client->ps.blend);
	}
	else if (ent->client->breather_framenum > level.framenum)
	{
		remaining = ent->client->breather_framenum - level.framenum;
		if (remaining == 30)	// beginning to fade
			gi.sound(ent, CHAN_ITEM, gi.soundindex("sound/items/airout.wav"), 1, ATTN_NORM, 0);
		if (remaining > 30 || (remaining & 4) )
			SV_AddBlend (0.4, 1, 0.4, 0.04, ent->client->ps.blend);
	}

	// add for damage
	if (ent->client->damage_alpha > 0)
		SV_AddBlend (ent->client->damage_blend[0],ent->client->damage_blend[1]
		,ent->client->damage_blend[2], ent->client->damage_alpha, ent->client->ps.blend);

	if (ent->client->bonus_alpha > 0)
		SV_AddBlend (0.85, 0.7, 0.3, ent->client->bonus_alpha, ent->client->ps.blend);

	// drop the damage value
	ent->client->damage_alpha -= 0.06;
	if (ent->client->damage_alpha < 0)
		ent->client->damage_alpha = 0;

	// drop the bonus value
	ent->client->bonus_alpha -= 0.1;
	if (ent->client->bonus_alpha < 0)
		ent->client->bonus_alpha = 0;
}

/*
=================
P_FootstepEvent
=================
*/
static void P_FootstepEvent ( edict_t *ent )
{
	if ( ent->s.event ) {
		return;
	}

	if ( pmtrace.surface->flags & SURF_NOSTEPS ) {
		return;
	}

	if ( pmtrace.surface->flags & SURF_METALSTEPS ) {
		ent->s.event = EV_FOOTSTEP_METAL;
	} else {
		ent->s.event = EV_FOOTSTEP;
	}
}

/*
=================
P_FallingDamage
=================
*/
void P_FallingDamage (edict_t *ent)
{
	float	delta;
	int		damage;
	vec3_t	dir;

	if (ent->s.modelindex != 255)
		return;		// not in the player model

 	if (ent->movetype == MOVETYPE_NOCLIP)
		return;

	if ((ent->client->oldvelocity[2] < 0) && (ent->velocity[2] > ent->client->oldvelocity[2]) && !ent->groundentity)
	{
		delta = ent->client->oldvelocity[2];
	}
	else
	{
		if (!ent->groundentity)
			return;
		delta = ent->velocity[2] - ent->client->oldvelocity[2];
	}
	delta = delta * delta * 0.0001;

//ZOID
	// never take damage if just release grapple or on grapple
	if (level.time - ent->client->ctf_grapplereleasetime <= FRAMETIME * 2 ||
		(ent->client->ctf_grapple &&
		ent->client->ctf_grapplestate > CTF_GRAPPLE_STATE_FLY))
		return;
//ZOID

	// scale delta if was pushed by jump pad
	if ( ent->client->jumppad_time && ent->client->jumppad_time < level.time ) {
		delta /= (1 + level.time - ent->client->jumppad_time) * 0.5;
		ent->client->jumppad_time = 0;
	}

	// never take falling damage if completely underwater
	if (ent->waterlevel == 3)
		return;
	if (ent->waterlevel == 2)
		delta *= 0.25;
	if (ent->waterlevel == 1)
		delta *= 0.5;

	if (delta < 1)
		return;

	if (delta < 15)
	{
		P_FootstepEvent ( ent );
		return;
	}

	// never take damage
	if ( pmtrace.surface->flags & SURF_NODAMAGE ) {
		if ( !ent->s.event )
			ent->s.event = EV_FALL_SHORT;
		return;
	}

	ent->client->fall_value = delta*0.5;
	if (ent->client->fall_value > 40)
		ent->client->fall_value = 40;
	ent->client->fall_time = level.time + FALL_TIME;

	if (delta > 30)
	{
		if (ent->health > 0)
		{
			if (delta >= 55)
				ent->s.event = EV_FALL_FAR;
			else
				ent->s.event = EV_FALL_MEDIUM;
		}
		ent->pain_debounce_time = level.time;	// no normal pain sound
		damage = (delta-30)/2;
		if (damage < 1)
			damage = 1;
		VectorSet (dir, 0, 0, 1);

		if (!deathmatch->value || !((int)dmflags->value & DF_NO_FALLING) )
			T_Damage (ent, world, world, dir, ent->s.origin, vec3_origin, damage, 0, 0, MOD_FALLING);
	}
	else
	{
		if ( !ent->s.event )		// lower priority than jump for example
			ent->s.event = EV_FALL_SHORT;
		return;
	}
}



/*
=============
P_WorldEffects
=============
*/
void P_WorldEffects (void)
{
	qboolean	breather;
	qboolean	envirosuit;
	int			waterlevel, old_waterlevel;

	if (current_player->movetype == MOVETYPE_NOCLIP)
	{
		current_player->air_finished = level.time + 12;	// don't need air
		return;
	}

	waterlevel = current_player->waterlevel;
	old_waterlevel = current_client->old_waterlevel;
	current_client->old_waterlevel = waterlevel;

	breather = current_client->breather_framenum > level.framenum;
	envirosuit = current_client->enviro_framenum > level.framenum;

	//
	// if just entered a water volume, play a sound
	//
	if (!old_waterlevel && waterlevel)
	{
		PlayerNoise(current_player, current_player->s.origin, PNOISE_SELF);
		if (current_player->watertype & CONTENTS_LAVA)
			gi.sound (current_player, CHAN_BODY, gi.soundindex("sound/player/lava_in.wav"), 1, ATTN_NORM, 0);
		else if (current_player->watertype & CONTENTS_SLIME)
			gi.sound (current_player, CHAN_BODY, gi.soundindex("sound/player/watr_in.wav"), 1, ATTN_NORM, 0);
		else if (current_player->watertype & CONTENTS_WATER)
			gi.sound (current_player, CHAN_BODY, gi.soundindex("sound/player/watr_in.wav"), 1, ATTN_NORM, 0);
		current_player->flags |= FL_INWATER;

		// clear damage_debounce, so the pain sound will play immediately
		current_player->damage_debounce_time = level.time - 1;
	}

	//
	// if just completely exited a water volume, play a sound
	//
	if (old_waterlevel && ! waterlevel)
	{
		PlayerNoise(current_player, current_player->s.origin, PNOISE_SELF);
		gi.sound (current_player, CHAN_BODY, gi.soundindex("sound/player/watr_out.wav"), 1, ATTN_NORM, 0);
		current_player->flags &= ~FL_INWATER;
	}

	//
	// check for head just going under water
	//
	if (old_waterlevel != 3 && waterlevel == 3)
	{
		gi.sound (current_player, CHAN_BODY, gi.soundindex("sound/player/watr_un.wav"), 1, ATTN_NORM, 0);
	}

	//
	// check for head just coming out of water
	//
	if (old_waterlevel == 3 && waterlevel != 3)
	{
		if (current_player->air_finished < level.time)
		{	// gasp for air
			gi.sound (current_player, CHAN_VOICE, gi.soundindex("sound/player/gasp1.wav"), 1, ATTN_NORM, 0);
			PlayerNoise(current_player, current_player->s.origin, PNOISE_SELF);
		}
		else  if (current_player->air_finished < level.time + 11)
		{	// just break surface
			gi.sound (current_player, CHAN_VOICE, gi.soundindex("sound/player/gasp2.wav"), 1, ATTN_NORM, 0);
		}
	}

	//
	// check for drowning
	//
	if (waterlevel == 3)
	{
		// breather or envirosuit give air
		if (breather || envirosuit)
		{
			current_player->air_finished = level.time + 10;

			if (((int)(current_client->breather_framenum - level.framenum) % 25) == 0)
			{
				if (!current_client->breather_sound)
					gi.sound (current_player, CHAN_AUTO, gi.soundindex("sound/player/u_breath1.wav"), 1, ATTN_NORM, 0);
				else
					gi.sound (current_player, CHAN_AUTO, gi.soundindex("sound/player/u_breath2.wav"), 1, ATTN_NORM, 0);

				current_client->breather_sound ^= 1;
				PlayerNoise(current_player, current_player->s.origin, PNOISE_SELF);

				// FIXME: release a bubble?
			}
		}

		// if out of air, start drowning
		if (current_player->air_finished < level.time)
		{	// drown!
			if (current_player->client->next_drown_time < level.time
				&& current_player->health > 0)
			{
				current_player->client->next_drown_time = level.time + 1;

				// take more damage the longer underwater
				current_player->dmg += 2;
				if (current_player->dmg > 15)
					current_player->dmg = 15;

				// play a gurp sound instead of a normal pain sound
				if (current_player->health <= current_player->dmg)
					gi.sound (current_player, CHAN_VOICE, gi.soundindex("sound/player/drown1.wav"), 1, ATTN_NORM, 0);
				else if (rand()&1)
					gi.sound (current_player, CHAN_VOICE, gi.soundindex("*gurp1.wav"), 1, ATTN_NORM, 0);
				else
					gi.sound (current_player, CHAN_VOICE, gi.soundindex("*gurp2.wav"), 1, ATTN_NORM, 0);

				current_player->pain_debounce_time = level.time;

				T_Damage (current_player, world, world, vec3_origin, current_player->s.origin, vec3_origin, current_player->dmg, 0, DAMAGE_NO_ARMOR, MOD_WATER);
			}
		}
	}
	else
	{
		current_player->air_finished = level.time + 12;
		current_player->dmg = 2;
	}

	//
	// check for sizzle damage
	//
	if (waterlevel && (current_player->watertype&(CONTENTS_LAVA|CONTENTS_SLIME)) )
	{
		if (current_player->watertype & CONTENTS_LAVA)
		{
			if (current_player->health > 0
				&& current_player->pain_debounce_time <= level.time
				&& current_client->invincible_framenum < level.framenum)
			{
				if (rand()&1)
					gi.sound (current_player, CHAN_VOICE, gi.soundindex("sound/player/burn1.wav"), 1, ATTN_NORM, 0);
				else
					gi.sound (current_player, CHAN_VOICE, gi.soundindex("sound/player/burn2.wav"), 1, ATTN_NORM, 0);
				current_player->pain_debounce_time = level.time + 1;
			}

			if (envirosuit)	// take 1/3 damage with envirosuit
				T_Damage (current_player, world, world, vec3_origin, current_player->s.origin, vec3_origin, 1*waterlevel, 0, 0, MOD_LAVA);
			else
				T_Damage (current_player, world, world, vec3_origin, current_player->s.origin, vec3_origin, 3*waterlevel, 0, 0, MOD_LAVA);
		}

		if (current_player->watertype & CONTENTS_SLIME)
		{
			if (!envirosuit)
			{	// no damage from slime with envirosuit
				T_Damage (current_player, world, world, vec3_origin, current_player->s.origin, vec3_origin, 1*waterlevel, 0, 0, MOD_SLIME);
			}
		}
	}
}


/*
===============
G_SetClientEffects
===============
*/
void G_SetClientEffects (edict_t *ent)
{
	int		pa_type;
	int		remaining;

	ent->s.effects = 0;
	ent->s.renderfx = 0;

	if (ent->health <= 0 || level.intermissiontime)
		return;

	if (ent->powerarmor_time > level.time)
	{
		pa_type = PowerArmorType (ent);
		if (pa_type == POWER_ARMOR_SCREEN)
		{
			ent->s.effects |= EF_POWERSCREEN;
		}
		else if (pa_type == POWER_ARMOR_SHIELD)
		{
			ent->s.renderfx |= RF_SHELL_GREEN;
		}
	}

//ZOID
	CTFEffects(ent);
//ZOID

	if (ent->client->quad_framenum > level.framenum)
	{
		remaining = ent->client->quad_framenum - level.framenum;
		if (remaining > 30 || (remaining & 4) )
//			ent->s.effects |= EF_QUAD;
			CTFSetPowerUpEffect(ent, EF_QUAD);
	}

	if (ent->client->invincible_framenum > level.framenum)
	{
		remaining = ent->client->invincible_framenum - level.framenum;
		if (remaining > 30 || (remaining & 4) )
//			ent->s.effects |= EF_PENT;
			CTFSetPowerUpEffect(ent, EF_PENT);
	}

	// show cheaters!!!
	if (!sv_godall->value && ent->flags & FL_GODMODE)
	{
		ent->s.renderfx |= (RF_SHELL_RED|RF_SHELL_GREEN|RF_SHELL_BLUE);
	}

}


/*
===============
G_SetClientEvent
===============
*/
void G_SetClientEvent (edict_t *ent)
{
	if ( ent->s.event )
		return;

	if ( (int)(current_client->bobtime+bobmove) != bobcycle ) {
		if ( ent->groundentity && (xyspeed > 225) && (ent->waterlevel == 0) ) {
			P_FootstepEvent ( ent );
		} else if ( ent->waterlevel == 1 ) {
			ent->s.event = EV_FOOTSPLASH;
		} else if ( ent->waterlevel == 2 ) {
			ent->s.event = EV_SWIM;
		}
	}
}

/*
===============
G_SetClientSound
===============
*/
void G_SetClientSound (edict_t *ent)
{
	char	*weap;

	if (ent->client->resp.game_helpchanged != game.helpchanged)
	{
		ent->client->resp.game_helpchanged = game.helpchanged;
		ent->client->resp.helpchanged = 1;
	}

	// help beep (no more than three times)
	if (ent->client->resp.helpchanged && ent->client->resp.helpchanged <= 3 && !(level.framenum&63) )
	{
		ent->client->resp.helpchanged++;
		gi.sound (ent, CHAN_VOICE, gi.soundindex ("sound/misc/pc_up.wav"), 1, ATTN_STATIC, 0);
	}


	if (ent->client->pers.weapon)
		weap = ent->client->pers.weapon->classname;
	else
		weap = "";

	if (ent->waterlevel && (ent->watertype&(CONTENTS_LAVA|CONTENTS_SLIME)) )
		ent->s.sound = snd_fry;
	else if (strcmp(weap, "weapon_railgun") == 0)
		ent->s.sound = gi.soundindex("sound/weapons/railgun/rg_hum.wav");
	else if (strcmp(weap, "weapon_bfg") == 0)
		ent->s.sound = gi.soundindex("sound/weapons/bfg_hum.wav");
	else if (ent->client->weapon_sound)
		ent->s.sound = ent->client->weapon_sound;
	else
		ent->s.sound = 0;
}

/*
=================================
 G_SetNewAnim - JAL splitmodels
=================================
*/
void G_SetNewAnim (edict_t *ent, qboolean updatetorso, qboolean updatelegs)
{
	gclient_t	*client;

	client = ent->client;

	if (ent->client->anim_priority == ANIM_DEATH)
		return;

	//SWIMMING
	if (client->anim_swim)
	{

		if (client->anim_duck)
		{
			//JAL FIXME: Crouching underwater really makes any sense?

			if (xyspeed){
				//CROUCHED move
				if (updatetorso)
					client->anim_upper = TORSO_CRWALK;

				if (updatelegs)
					client->anim_lower = LEGS_CRWALK;

			} else {
				//CROUCHED stand
				if (updatetorso)
					client->anim_upper = TORSO_CRSTAND;

				if (updatelegs)
					client->anim_lower = LEGS_CRSTAND;
			}

		} else {

			if (updatetorso)
				client->anim_upper = TORSO_SWIM;

			if (updatelegs)
			{
				if (client->anim_moveflags & ANIMMOVE_FRONT)
				{
					client->anim_lower = LEGS_SWIMFWD;

				} else
					client->anim_lower = LEGS_SWIM;
			}
		}
	}
	//is JUMP
	else if (client->anim_jump){

		if (client->ctf_grapple)
		{
			//GRAPPLE
			if (updatetorso)
				client->anim_upper = TORSO_STAND;

			if (updatelegs)
				client->anim_lower = LEGS_JUMP3;

		} else {

			//JUMP
			if (updatetorso)
				client->anim_upper = TORSO_RUN;

			if (updatelegs)
			{
				if (client->anim_jump_style == 1){
					client->anim_lower = LEGS_JUMP1;
				}
				else if (client->anim_jump_style == 2){
					client->anim_lower = LEGS_JUMP2;
				}
				else
					client->anim_lower = LEGS_JUMP3;
			}
		}
	}
	// RUN
	else if (client->anim_run)
	{
		if (client->anim_duck)
		{
			//CROUCHED move	(newframes)
			if (updatetorso)
				client->anim_upper = TORSO_CRWALK;

			if (updatelegs)
				client->anim_lower = LEGS_CRWALK;

		} else {

			//STANDUP
			if (client->anim_moveflags & ANIMMOVE_WALK)
			{
				//WALK
				if (updatetorso)
					client->anim_upper = TORSO_STAND;

				if (updatelegs)
				{
					//front/backward has priority over side movements
					if (client->anim_moveflags & ANIMMOVE_FRONT){
						client->anim_lower = LEGS_WALKFWD;

					} else if (client->anim_moveflags & ANIMMOVE_BACK){
						client->anim_lower = LEGS_WALKBACK;

					} else if (client->anim_moveflags & ANIMMOVE_RIGHT){
						client->anim_lower = LEGS_WALKRIGHT;

					} else if (client->anim_moveflags & ANIMMOVE_LEFT){
						client->anim_lower = LEGS_WALKLEFT;

					} else 	//is moving by inertia
						client->anim_lower = LEGS_WALKFWD;
				}

			} else {

				//RUN
				if (updatetorso)
					client->anim_upper = TORSO_RUN;

				if (updatelegs)
				{
					//front/backward has priority over side movements
					if (client->anim_moveflags & ANIMMOVE_FRONT){
						client->anim_lower = LEGS_RUNFWD;

					} else if (client->anim_moveflags & ANIMMOVE_BACK){
						client->anim_lower = LEGS_RUNBACK;

					} else if (client->anim_moveflags & ANIMMOVE_RIGHT){
						client->anim_lower = LEGS_RUNRIGHT;

					} else if (client->anim_moveflags & ANIMMOVE_LEFT){
						client->anim_lower = LEGS_RUNLEFT;

					} else 	//is moving by inertia
						client->anim_lower = LEGS_WALKFWD;
				}
			}
		}
	}
	else
	{	// STAND
		if (client->anim_duck)
		{
			//CROUCHED stand
			if (updatetorso)
				client->anim_upper = TORSO_CRSTAND;

			if (updatelegs)
				client->anim_lower = LEGS_CRSTAND;

		} else {

			//STAND STANDUP
			if (updatetorso)
				client->anim_upper = TORSO_STAND;

			if (updatelegs)
				client->anim_lower = LEGS_STAND;
		}
	}
}

/*
===================
  Anim_CheckJump - JAL SPLITMODELS
  Manages the jumping animation phases. First phase is the player going up.
  Second: It checks the floor one step below the player, when it detects
  floor it launches the landing animation.
===================
*/
void Anim_CheckJump (edict_t *ent)
{
	vec3_t		point;
	trace_t		trace;

	//not jumping anymore?
	if (ent->groundentity)
	{
		ent->client->anim_jump_thunk = qfalse;
		ent->client->anim_jump = qfalse;
		return;
	}

	//if not falling we don't need to continue
	if (ent->client->ps.pmove.velocity[2] > -80)
		return;

	//PRE-STEP TO LAND
	if (ent->client->anim_jump_prestep)	//I already did.
		return;

	//determine a point below scaled by velocity
	point[0] = ent->s.origin[0];
	point[1] = ent->s.origin[1];
	point[2] = ent->s.origin[2] + ((0.025 * ent->client->ps.pmove.velocity[2]));

	//trace to point
	trace = gi.trace (ent->s.origin, ent->mins, ent->maxs, point, ent, MASK_PLAYERSOLID);

	if ( !trace.ent || (trace.plane.normal[2] < 0.7 && !trace.startsolid) )
	{
		return;
	}
	else
	{
		//found solid. Get ready to land!
		if (ent->client->anim_jump_style == 1){
			ent->client->anim_lower = LEGS_JUMP1ST;

		} else if (ent->client->anim_jump_style == 2){
			ent->client->anim_lower = LEGS_JUMP2ST;

		} else
			ent->client->anim_lower = LEGS_JUMP3ST;

		//set as done
		ent->client->anim_jump_prestep = qtrue;
	}
}

/*
===================
  AnimIsStep - JAL SPLITMODELS
  Checks the floor one step below the player. Used to detect
  if the player is really falling or just walking down a stair.
===================
*/
#define STEPSIZE 18
qboolean Anim_IsStep (edict_t *ent)
{
	vec3_t		point;
	trace_t		trace;

	//determine a point below
	point[0] = ent->s.origin[0];
	point[1] = ent->s.origin[1];
	point[2] = ent->s.origin[2] - (1.6*STEPSIZE);

	//trace to point
	trace = gi.trace (ent->s.origin, ent->mins, ent->maxs, point, ent, MASK_PLAYERSOLID);

	if ( !trace.ent || (trace.plane.normal[2] < 0.7 && !trace.startsolid) )
	{
		return qfalse;
	}

	//found solid.
	return qtrue;
}
/*
===================
  Anim_IsSwim - JAL SPLITMODELS
  Categorize waterlevel for swim animation.
===================
*/
qboolean Anim_IsSwim (edict_t *ent){

	if (ent->waterlevel > 2)
		return qtrue;

	if (ent->waterlevel && !ent->groundentity){

		if (!Anim_IsStep(ent))
			return qtrue;
	}

	return qfalse;
}

/*
=================
G_SetClientFrame - JAL SPLITMODELS
=================
*/
void G_SetClientFrame (edict_t *ent)
{
	gclient_t	*client;
	qboolean	duck, run, swim, updatetorso, updatelegs;
	int			upperfinal, lowerfinal, headfinal;

	if ((!ent->client) || (ent->s.modelindex != 255) || (ent->svflags & SVF_NOCLIENT))
		return;

	client = ent->client;

		//ANIM_DEAD EXCEPTION
	if (ent->client->anim_priority == ANIM_DEATH){

		//count down
		if (client->anim_event_time > 0)
			client->anim_event_time--;

		//just keep the values
		upperfinal = (int)ent->s.frame & 0x3F;
		lowerfinal = (int)(ent->s.frame >>6) & 0x3F;
		headfinal = (int)(ent->s.frame >>12) & 0xF;

	} else if (ent->client->anim_priority == ANIM_CINEMATIC){

		//EXTERNAL ANIMATIONS:
		//just keep the values and don't count down.
		//it will never autoexit this priority, nor
		//change the introduced animation, you will
		//have to make it leave from the outside.

		client->anim_event_time = 1;

		upperfinal = (int)ent->s.frame & 0x3F;
		lowerfinal = (int)(ent->s.frame >>6) & 0x3F;
		headfinal = (int)(ent->s.frame >>12) & 0xF;

	} else {

		//PLAYER ANIMATIONS

		updatetorso = updatelegs = qfalse;

		//Prepare the checking values
		if (client->ps.pmove.pm_flags & PMF_DUCKED)
			duck = qtrue;
		else
			duck = qfalse;

		if (xyspeed)
			run = qtrue;
		else
			run = qfalse;

		swim = Anim_IsSwim(ent);

		//jump is special
		if (client->anim_jump && (ent->groundentity || swim))//JUMP stop
		{
			client->anim_jump = qfalse;
			client->anim_jump_thunk = qfalse;

		} else if (!ent->groundentity && !client->anim_jump && !swim){ //falling

			if (!Anim_IsStep(ent)){
				client->anim_jump_style = 0;
				client->anim_jump = qtrue;
				updatelegs = qtrue;
			}
		}
		//jumping thinking
		if (client->anim_jump)
			Anim_CheckJump(ent);


		//---New animation?

		//Is swimming and wasn't
		if (swim != client->anim_swim){
			updatetorso = qtrue;
			updatelegs = qtrue;
		}
		//jump is ordered to refresh by setting anim_jump_thunk as false.
		else if (client->anim_jump_thunk == qfalse){
			updatetorso = qtrue;
			updatelegs = qtrue;
			//client->anim_jump_thunk = qtrue;
			client->anim_jump_prestep = qfalse;
		}
		// check for stand/duck transition
		else if (duck != client->anim_duck){
			updatetorso = qtrue;
			updatelegs = qtrue;
		}
		//check for stop/go transition
		else if (!client->anim_jump && (run != client->anim_run
			|| ((client->anim_moveflags & ANIMMOVE_WALK) != (client->anim_oldmoveflags & ANIMMOVE_WALK))
			)){
			updatetorso = qtrue;
			updatelegs = qtrue;
		}

		//LEGS only

		//player direction has changed, legs may change.
		if (client->anim_oldmoveflags != client->anim_moveflags && !client->anim_jump)
		{
			if (((client->anim_moveflags & ANIMMOVE_FRONT) && (client->anim_oldmoveflags & ANIMMOVE_FRONT)) || ((client->anim_moveflags & ANIMMOVE_BACK) && (client->anim_oldmoveflags & ANIMMOVE_BACK)))
			{
				//erm... fixme

			} else {
				updatelegs = qtrue;
			}
		}

		//Set the new animation
		if (updatetorso || updatelegs)
		{
			client->anim_jump_thunk = qtrue;
			client->anim_duck = duck;
			client->anim_run = run;
			client->anim_swim = swim;
			G_SetNewAnim (ent, updatetorso, updatelegs);
		}


		//---------------------------
		//Final animation management.
		//---------------------------

		//Update torso
		if (ent->client->anim_priority > ANIM_BASIC)
		{
			client->anim_event_time--;
			upperfinal = ent->s.frame & 0x3F;

			//event finished:
			if (client->anim_event_time <= 0)
				ent->client->anim_priority = ANIM_BASIC;
		}
		else
			upperfinal = client->anim_upper;

		//update legs
		lowerfinal = client->anim_lower;

		//there aren't any head animations.
		headfinal = ent->client->anim_head = 0;
	}

	//send only if it's a new animation.
	//set 'lastsent' fields as 0 to relaunch the same animation
	ent->s.frame = 0;

	if (upperfinal && upperfinal != client->upper_lastsent){
		ent->s.frame |= upperfinal & 0x3F;
		client->upper_lastsent = upperfinal;
	}
	if (lowerfinal && lowerfinal != client->lower_lastsent){
		ent->s.frame |= (lowerfinal & 0x3F)<<6;
		client->lower_lastsent = lowerfinal;
	}
	if (headfinal && headfinal != client->head_lastsent){
		ent->s.frame |= (headfinal & 0xF)<<12;
		client->head_lastsent = headfinal;
	}

	//ent->s.frame = ((uppersend & 0x3F)|(lowersend & 0x3F)<<6|(headsend & 0xF)<<12);

	//store oldmoveflags for next round.
	ent->client->anim_oldmoveflags = ent->client->anim_moveflags;
}

/*
=================
ClientEndServerFrame

Called for each player at the end of the server frame
and right after spawning
=================
*/
void ClientEndServerFrame (edict_t *ent)
{
	float	bobtime;
	int		i, j;//CHASEHACK & splitmodels
	vec3_t dir;

	current_player = ent;
	current_client = ent->client;

	//
	// If the origin or velocity have changed since ClientThink(),
	// update the pmove values.  This will happen when the client
	// is pushed by a bmodel or kicked by an explosion.
	//
	// If it wasn't updated here, the view position would lag a frame
	// behind the body position when pushed -- "sinking into plats"
	//
	for (i=0 ; i<3 ; i++)
	{
		current_client->ps.pmove.origin[i] = ent->s.origin[i]*8.0;
		current_client->ps.pmove.velocity[i] = ent->velocity[i]*8.0;
	}

	VectorSet ( dir, ent->s.origin[0], ent->s.origin[1], ent->s.origin[2] - 0.25 );

	if ( ent->health > 0 ) {
		if ( ent->groundentity ) {
			pmtrace = gi.trace ( ent->s.origin, ent->mins, ent->maxs, dir, ent, MASK_PLAYERSOLID );
		} else {
			pmtrace = gi.trace ( ent->s.origin, ent->mins, ent->maxs, ent->s.origin, ent, MASK_PLAYERSOLID );
		}
	} else {
		pmtrace = gi.trace ( ent->s.origin, ent->mins, ent->maxs, ent->s.origin, ent, MASK_DEADSOLID );
	}

	//
	// If the end of unit layout is displayed, don't give
	// the player any normal movement attributes
	//
	if (level.intermissiontime)
	{
		// FIXME: add view drifting here?
		current_client->ps.blend[3] = 0;
		current_client->ps.fov = 90;
		G_SetStats (ent);
		return;
	}

	AngleVectors (ent->client->v_angle, forward, right, up);

	// burn from lava, etc
	P_WorldEffects ();

	//
	// set model angles from view angles so other things in
	// the world can tell which direction you are looking
	//
	if (ent->client->v_angle[PITCH] > 180)
		ent->s.angles[PITCH] = (-360 + ent->client->v_angle[PITCH])/3; //qbism FIXME: bend torso
	else
		ent->s.angles[PITCH] = ent->client->v_angle[PITCH]/3; //qbism FIXME: bend torso
	ent->s.angles[YAW] = ent->client->v_angle[YAW];
	ent->s.angles[ROLL] = 0;
	ent->s.angles[ROLL] = SV_CalcRoll (ent->velocity)*4;

	//
	// calculate speed and cycle to be used for
	// all cyclic walking effects
	//
	xyspeed = sqrt(ent->velocity[0]*ent->velocity[0] + ent->velocity[1]*ent->velocity[1]);

	if (xyspeed < 5)
	{
		bobmove = 0;
		current_client->bobtime = 0;	// start at beginning of cycle again
	}
	else if (ent->groundentity)
	{	// so bobbing only cycles when on ground
		if (xyspeed > 210)
			bobmove = 0.25;
		else if (xyspeed > 100)
			bobmove = 0.125;
		else
			bobmove = 0.0625;
	}

	bobtime = (current_client->bobtime += bobmove);

	if (current_client->ps.pmove.pm_flags & PMF_DUCKED)
		bobtime *= 4;

	bobcycle = (int)bobtime;
	bobfracsin = fabs(sin(bobtime*M_PI));

	// detect hitting the floor
	P_FallingDamage (ent);

	// apply all the damage taken this frame
	P_DamageFeedback (ent);

	// determine the view offsets
	SV_CalcViewOffset (ent);

	// determine the gun offsets
	SV_CalcGunOffset (ent);

	// determine the full screen color blend
	// must be after viewoffset, so eye contents can be
	// accurately determined
	// FIXME: with client prediction, the contents
	// should be determined by the client
	SV_CalcBlend (ent);

//ZOID
	if (!ent->client->chase_target)
//ZOID
		G_SetStats (ent);



//ZOID
//CHASEHACKv2 & splitmodels: jal [begin]

	//update chasecam follower stats & 1st view effects
	for (i = 1; i <= maxclients->value; i++) {
		edict_t *e = g_edicts + i;
		if (!e->inuse || e->client->chase_target != ent)
			continue;

		//follower stats
		memcpy(e->client->ps.stats, ent->client->ps.stats, sizeof(ent->client->ps.stats));
		e->client->ps.stats[STAT_LAYOUTS] = 1;

		if (e->client->chase_mode == 1){//in eyes

			//KickAngles
			VectorCopy(ent->client->ps.viewoffset, e->client->ps.viewoffset);
			VectorCopy (ent->client->ps.viewangles, e->client->ps.viewangles);
			VectorCopy(ent->client->ps.gunangles, e->client->ps.gunangles);
			VectorCopy(ent->client->ps.kick_angles, e->client->ps.kick_angles);
			e->viewheight = ent->viewheight;
			//Blends
			for (j = 0; j<4 ; j++)
				e->client->ps.blend[j] = ent->client->ps.blend[j];
		}
		break;
	}
//jal [end]
//ZOID

	G_SetClientEvent (ent);

	G_SetClientEffects (ent);

	G_SetClientSound (ent);

	G_SetClientFrame (ent);

	VectorCopy (ent->velocity, ent->client->oldvelocity);
	VectorCopy (ent->client->ps.viewangles, ent->client->oldviewangles);

	// clear weapon kicks
	VectorClear (ent->client->kick_origin);
	VectorClear (ent->client->kick_angles);

	// if the scoreboard is up, update it
	if (ent->client->showscores && !(level.framenum & 31) )
	{
//ZOID
		if (ent->client->menu) {
			PMenu_Do_Update(ent);
			ent->client->menudirty = qfalse;
			ent->client->menutime = level.time;
		} else
//ZOID
			DeathmatchScoreboardMessage (ent, ent->enemy);
		gi.unicast (ent, qfalse);
	}
}

