/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cmodel.c -- model loading

#include "qcommon.h"

#define MAX_CM_AREAS			(MAX_MAP_AREAS)
#define MAX_CM_PORTALS			(MAX_EDICTS)
#define MAX_CM_AREAPORTALS		(MAX_CM_PORTALS << 1)
#define MAX_CM_BRUSHSIDES		(MAX_MAP_BRUSHSIDES << 1)
#define MAX_CM_SHADERS			(MAX_MAP_SHADERS)
#define MAX_CM_PLANES			(MAX_MAP_PLANES << 1)
#define MAX_CM_NODES			(MAX_MAP_NODES)
#define MAX_CM_LEAFS			(MAX_MAP_LEAFS)
#define MAX_CM_LEAFBRUSHES		(MAX_MAP_LEAFBRUSHES)
#define MAX_CM_MODELS			(MAX_MAP_MODELS)
#define MAX_CM_BRUSHES			(MAX_MAP_BRUSHES << 1)
#define MAX_CM_VISIBILITY		(MAX_MAP_VISIBILITY)
#define MAX_CM_FACES			(MAX_MAP_FACES)
#define MAX_CM_LEAFFACES		(MAX_MAP_LEAFFACES)
#define MAX_CM_VERTEXES			(MAX_MAP_VERTEXES)
#define MAX_CM_PATCHES			(0x10000)
#define MAX_CM_PATCH_VERTS		(4096)
#define MAX_CM_ENTSTRING		(MAX_MAP_ENTSTRING)

#define CM_PLANES_HASH			(1024)
#define CM_SUBDIV_LEVEL			(15)

//qbism LCCFIX moved most structs to qcommon.h


// each area has a list of portals that lead into other areas
// when portals are closed, other areas may not be visible or
// hearable even if the vis info says that it should be
typedef struct
{
	int		portalnum;
	int		otherarea;
} careaportal_t;

typedef struct
{
	int		numareaportals;
	int		areaportals[MAX_CM_AREAPORTALS];
	int		floodnum;			// if two areas have equal floodnums, they are connected
	int		floodvalid;
} carea_t;

int			checkcount;

char		map_name[MAX_QPATH];

int			numbrushsides;
cbrushside_t map_brushsides[MAX_CM_BRUSHSIDES+6];	// extra for box hull

int			numshaderrefs;
csurface_t	map_surfaces[MAX_CM_SHADERS];

int			numplanes;
cplane_t	map_planes[MAX_CM_PLANES+6];

int			numnodes;
cnode_t		map_nodes[MAX_CM_NODES];

int			numleafs = 1;	// allow leaf funcs to be called without a map
cleaf_t		map_leafs[MAX_CM_LEAFS];

int			numleafbrushes;
int			map_leafbrushes[MAX_CM_LEAFBRUSHES];

int			numcmodels;
cmodel_t	map_cmodels[MAX_CM_MODELS+1];			// extra for box hull

int			numbrushes;
cbrush_t	map_brushes[MAX_CM_BRUSHES+1];			// extra for box hull

cface_t		map_faces[MAX_CM_FACES];
int			numfaces;

int			map_leafpatches[MAX_CM_LEAFFACES];
int			numleafpatches;

vec4_t		*map_verts;								// this will be freed
int			numvertexes;

int			*map_leaffaces;							// this will be freed
int			numleaffaces;

planehash_t *map_planeshash[CM_PLANES_HASH];

qbyte		map_hearability[MAX_CM_VISIBILITY];

qbyte		map_visibility[MAX_CM_VISIBILITY];
dvis_t		*map_pvs = (dvis_t *)map_visibility;

dvis_t		*map_phs = (dvis_t *)map_hearability;

qbyte		nullrow[MAX_CM_LEAFS/8];

int			numentitychars;
char		map_entitystring[MAX_CM_ENTSTRING];

int			numareaportals = 1;
careaportal_t map_areaportals[MAX_CM_AREAPORTALS];

int			numareas = 1;
carea_t		map_areas[MAX_CM_AREAS];

csurface_t	nullsurface;

int			floodvalid;

qboolean	portalopen[MAX_CM_PORTALS];

cvar_t		*cm_noAreas;
cvar_t		*cm_noCurves;

void	CM_InitBoxHull (void);
void	FloodAreaConnections (void);

int		c_pointcontents;
int		c_traces, c_brush_traces;

qbyte	*cmod_base;

/*
===============================================================================

					UTILITY FUNCTIONS

===============================================================================
*/

#define	PLANE_NORMAL_EPSILON	0.00001
#define	PLANE_DIST_EPSILON		0.01

/*
=================
CM_ComparePlanes
=================
*/
static qboolean CM_ComparePlanes ( const cplane_t *p, const vec3_t normal, const vec_t dist )
{
	if (fabs (p->normal[0] - normal[0]) < PLANE_NORMAL_EPSILON
		&& fabs (p->normal[1] - normal[1]) < PLANE_NORMAL_EPSILON
		&& fabs (p->normal[2] - normal[2]) < PLANE_NORMAL_EPSILON
		&& fabs (p->dist - dist) < PLANE_DIST_EPSILON )
		return qtrue;

	return qfalse;
}

/*
==============
CM_SnapVector
==============
*/
static void CM_SnapVector ( vec3_t normal )
{
	int		i;

	for ( i = 0; i < 3; i++ ) {
		if ( fabs (normal[i] - 1) < PLANE_NORMAL_EPSILON ) {
			VectorClear ( normal );
			normal[i] = 1;
			break;
		}
		if ( fabs (normal[i] - -1) < PLANE_NORMAL_EPSILON ) {
			VectorClear ( normal );
			normal[i] = -1;
			break;
		}
	}
}

/*
==============
CM_SnapPlane
==============
*/
static void CM_SnapPlane ( vec3_t normal, vec_t *dist )
{
	CM_SnapVector ( normal );

	if ( fabs (*dist - Q_rint (*dist)) < PLANE_DIST_EPSILON ) {
		*dist = Q_rint ( *dist );
	}
}

/*
=================
CM_HashKeyForPlane
=================
*/
static inline int CM_HashKeyForPlane ( float dist )
{
	return ((int)( fabs (dist) ) / 8) & (CM_PLANES_HASH - 1);
}

/*
=================
CM_AddPlaneToHash
=================
*/
static cplane_t *CM_AddPlaneToHash ( int key, cplane_t *plane )
{
	planehash_t *hash;

	hash = Z_Malloc ( sizeof(planehash_t) );
	hash->plane = plane;
	hash->hashNext = map_planeshash[key];
	map_planeshash[key] = hash;

	return plane;
}

/*
=================
CM_FreePlanesHash
=================
*/
static void CM_FreePlanesHash (void)
{
	int i;
	planehash_t *hash, *hashNext;

	for ( i = 0; i < CM_PLANES_HASH; i++ ) {
		hash = map_planeshash[i];

		for ( ; hash; hash = hashNext ) {
			hashNext = hash->hashNext;
			hash->hashNext = NULL;
			Z_Free ( hash );
		}
	}

	memset ( map_planeshash, 0, sizeof(map_planeshash) );
}

/*
=================
CM_AddPlane
=================
*/
static cplane_t *CM_AddPlane ( vec3_t normal, vec_t dist )
{
	int key;
	cplane_t *plane;
	planehash_t *hash;

	CM_SnapPlane ( normal, &dist );

	key = CM_HashKeyForPlane ( dist );
	for ( hash = map_planeshash[key]; hash; hash = hash->hashNext ) {
		if ( CM_ComparePlanes (hash->plane, normal, dist) ) {
			return hash->plane;
		}
	}

	if ( numplanes == MAX_CM_PLANES ) {
		Com_Error ( ERR_DROP, "CM_AddPlane: numplanes == MAX_CM_PLANES" );
	}

	plane = &map_planes[numplanes++];
	VectorCopy ( normal, plane->normal );
	plane->dist = dist;
	CategorizePlane ( plane );

	return CM_AddPlaneToHash ( key, plane );
}

/*
=================
CM_AddBrushside
=================
*/
static void CM_AddBrushside ( cbrush_t *brush, const cplane_t *plane, const csurface_t *surface )
{
	cbrushside_t *side;

	if ( numbrushsides == MAX_CM_BRUSHSIDES ) {
		Com_Error ( ERR_DROP, "CM_AddBrushside: numbrushsides == MAX_CM_BRUSHSIDES" );
	}

	side = &map_brushsides[numbrushsides++];
	side->plane = ( cplane_t * )plane;
	side->surface = ( csurface_t * )surface;
	brush->numsides++;
}

/*
===============================================================================

					PATCH LOADING

===============================================================================
*/

/*
=================
CM_CreateBrushFromPoints
=================
*/
static int CM_CreateBrushFromPoints ( cbrush_t *brush, vec3_t *verts, csurface_t *surface )
{
	int				i, j, k;
	int				axis, dir;
	vec3_t			mins, maxs;
	cbrushside_t	*s;
	vec3_t			normal;
	float			d, dist;
	cplane_t		mainplane;
	vec3_t			vec, vec2;

	// set default values for brush
	brush->numsides = 0;
	brush->firstbrushside = &map_brushsides[numbrushsides];
	brush->contents = surface->contents;

	// calculate plane for this triangle
	PlaneFromPoints ( verts, &mainplane );
	if ( CM_ComparePlanes (&mainplane, vec3_origin, mainplane.dist) ) {
		return 0;
	}
	CM_SnapPlane ( mainplane.normal, &mainplane.dist );

	// calculate mins & maxs
	ClearBounds ( mins, maxs );
	for ( i = 0; i < 3; i++ ) {
		AddPointToBounds ( verts[i], mins, maxs );
	}

	// add the axial planes
	for ( axis = 0; axis < 3; axis++ ) {
		for ( dir = -1; dir <= 1; dir += 2 ) {
			VectorClear ( normal );
			normal[axis] = dir;
			if (dir == 1)
				dist = maxs[axis];
			else
				dist = -mins[axis];

			if ( !CM_ComparePlanes (&mainplane, normal, dist) ) {
				CM_AddBrushside ( brush, CM_AddPlane (normal, dist), surface );
			}
		}
	}

	// add front plane
	CM_AddBrushside ( brush, CM_AddPlane (mainplane.normal, mainplane.dist), surface );

	// add the edge bevels
	for ( i = 0; i < 3; i++ ) {
		j = (i + 1) % 3;
		k = (i + 2) % 3;

		VectorSubtract ( verts[i], verts[j], vec );
		if ( VectorNormalize (vec) < 0.5 ) {
			continue;
		}

		CM_SnapVector ( vec );
		for ( j = 0; j < 3; j++ ) {
			if ( vec[j] == 1 || vec[j] == -1 ) {
				break;	// axial
			}
		}
		if ( j != 3 ) {
			continue;	// only test non-axial edges
		}

		// try the six possible slanted axials from this edge
		for ( axis = 0; axis < 3; axis++ ) {
			for ( dir = -1; dir <= 1; dir += 2 ) {
				// construct a plane
				VectorClear ( vec2 );
				vec2[axis] = dir;
				CrossProduct ( vec, vec2, normal );
				if ( VectorNormalize (normal) < 0.5 ) {
					continue;
				}

				dist = DotProduct ( verts[i], normal );

				s = brush->firstbrushside;
				for ( j = 0; j < brush->numsides; j++, s++ ) {
					// if this plane has allready been used, skip it
					if ( CM_ComparePlanes (s->plane, normal, dist) ) {
						break;
					}
				}

				if ( j != brush->numsides ) {
					continue;
				}

				// if third point of the triangle is
				// behind this plane, it is a proper edge bevel
				d = DotProduct ( verts[k], normal ) - dist;
				if ( d > 0.1 ) {
					continue;	// point in front: this plane isn't part of the outer hull
				}

				// add this plane
				CM_AddBrushside ( brush, CM_AddPlane (normal, dist), surface );
			}
		}
	}

	return brush->numsides;
}

/*
=================
CM_CreatePatch
=================
*/
static void CM_CreatePatch ( cface_t *face, int numverts, vec4_t *verts, int *patch_cp )
{
    int step[2], size[2], flat[2], i, u, v;
	vec4_t points[MAX_CM_PATCH_VERTS];
	vec3_t tverts[4], tverts2[4];
	cbrush_t *brush;

// find the degree of subdivision in the u and v directions
	Patch_GetFlatness ( CM_SUBDIV_LEVEL, verts, patch_cp, flat );

	step[0] = 1 << flat[0];
	step[1] = 1 << flat[1];
	size[0] = ( patch_cp[0] >> 1 ) * step[0] + 1;
	size[1] = ( patch_cp[1] >> 1 ) * step[1] + 1;

	if ( size[0] * size[1] > MAX_CM_PATCH_VERTS ) {
		Com_Error ( ERR_DROP, "CM_CreatePatch: patch has too many vertices" );
		return;
	}

// fill in
	Patch_Evaluate ( verts, patch_cp, step, points );

	face->brushes = brush = map_brushes + numbrushes;
	face->numbrushes = 0;
	ClearBounds ( face->mins, face->maxs );

// create a set of brushes
    for ( v = 0; v < size[1]-1; v++ ) {
		for ( u = 0; u < size[0]-1; u++ ) {
			if ( numbrushes >= MAX_CM_BRUSHES ) {
				Com_Error (ERR_DROP, "CM_CreatePatch: too many patch brushes");
			}

			i = v * size[0] + u;
			VectorCopy ( points[i], tverts[0] );
			VectorCopy ( points[i + size[0]], tverts[1] );
			VectorCopy ( points[i + 1], tverts[2] );
			VectorCopy ( points[i + size[0] + 1], tverts[3] );

			for ( i = 0; i < 4; i++ ) {
				AddPointToBounds ( tverts[i], face->mins, face->maxs );
			}

			// create two brushes
			if ( CM_CreateBrushFromPoints (brush, tverts, face->surface) ) {
				brush++; numbrushes++; face->numbrushes++;
			}

			VectorCopy ( tverts[2], tverts2[0] );
			VectorCopy ( tverts[1], tverts2[1] );
			VectorCopy ( tverts[3], tverts2[2] );

			if ( CM_CreateBrushFromPoints (brush, tverts2, face->surface) ) {
				brush++; numbrushes++; face->numbrushes++;
			}
		}
    }

	for ( i = 0; i < 3; i++ ) {
		// spread the mins / maxs by a pixel
		face->mins[i] -= 1;
		face->maxs[i] += 1;
	}
}


/*
===============================================================================

					MAP LOADING

===============================================================================
*/

/*
=================
CMod_LoadSurfaces
=================
*/
void CMod_LoadSurfaces (lump_t *l)
{
	dshaderref_t	*in;
	csurface_t		*out;
	int				i, count;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map with no shaders");
	else if (count > MAX_MAP_SHADERS)
		Com_Error (ERR_DROP, "Map has too many shaders");

	numshaderrefs = count;
	out = map_surfaces;

	for ( i=0 ; i<count ; i++, in++, out++ ) {
		out->flags = LittleLong ( in->flags );
		out->contents = LittleLong ( in->contents );
	}
}

/*
=================
CMod_LoadVertexes
=================
*/
void CMod_LoadVertexes (lump_t *l)
{
	dvertex_t	*in;
	vec4_t		*out;
	int			i, count, j;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "CMOD_LoadVertexes: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count > MAX_MAP_VERTEXES)
		Com_Error (ERR_DROP, "Map has too many vertexes");

	out = Z_Malloc ( count*sizeof(*out) );
	map_verts = out;
	numvertexes = count;

	for ( i=0 ; i<count ; i++, in++)
	{
		for ( j=0 ; j < 3 ; j++)
		{
			out[i][j] = LittleFloat ( in->point[j] );
		}
	}
}


/*
=================
CMod_LoadFaces
=================
*/
void CMod_LoadFaces (lump_t *l)
{
	dface_t		*in;
	cface_t		*out;
	int			i, count;
	int			numverts, firstvert;
	int			shadernum, patch_cp[2];

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count > MAX_MAP_FACES)
		Com_Error (ERR_DROP, "Map has too many faces");

	out = map_faces;
	numfaces = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->surface = NULL;
		out->brushes = NULL;
		out->numbrushes = 0;

		if ( LittleLong (in->facetype) != FACETYPE_PATCH ) {
			continue;
		}

		shadernum = LittleLong ( in->shadernum );
		if ( shadernum < 0 || shadernum >= numshaderrefs ) {
			continue;
		}
		out->surface = &map_surfaces[shadernum];
		if ( !out->surface->contents || (out->surface->flags & SURF_NONSOLID) ) {
			continue;
		}

		patch_cp[0] = LittleLong ( in->patch_cp[0] );
		patch_cp[1] = LittleLong ( in->patch_cp[1] );
		if ( patch_cp[0] <= 0 || patch_cp[1] <= 0 ) {
			continue;
		}

		numverts = LittleLong ( in->numverts );
		firstvert = LittleLong ( in->firstvert );
		if ( numverts <= 0 || firstvert < 0 || firstvert >= numvertexes ) {
			continue;
		}

		CM_CreatePatch ( out, numverts, map_verts + firstvert, patch_cp );
	}
}

/*
=================
CMod_LoadSubmodels
=================
*/
void CMod_LoadSubmodels (lump_t *l)
{
	dmodel_t	*in;
	cmodel_t	*out;
	int			i, j, count;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map with no models");
	else if (count > MAX_MAP_MODELS)
		Com_Error (ERR_DROP, "Map has too many models");

	numcmodels = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out = &map_cmodels[i];
		out->numfaces = LittleLong ( in->numfaces );
		out->firstface = LittleLong ( in->numfaces );
		out->numbrushes = LittleLong ( in->numbrushes );
		out->firstbrush = LittleLong ( in->firstbrush );

		for ( j = 0; j < 3; j++ ) {
			// spread the mins / maxs by a pixel
			out->mins[j] = LittleFloat (in->mins[j]) - 1;
			out->maxs[j] = LittleFloat (in->maxs[j]) + 1;
		}
	}
}

/*
=================
CMod_LoadNodes

=================
*/
void CMod_LoadNodes (lump_t *l)
{
	dnode_t		*in;
	cnode_t		*out;
	int			i, count;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map has no nodes");
	else if (count > MAX_MAP_NODES)
		Com_Error (ERR_DROP, "Map has too many nodes");

	out = map_nodes;
	numnodes = count;

	for (i=0 ; i<count ; i++, out++, in++)
	{
		out->plane = map_planes + LittleLong(in->planenum);
		out->children[0] = LittleLong (in->children[0]);
		out->children[1] = LittleLong (in->children[1]);
	}
}

/*
=================
CMod_LoadLeafFaces
=================
*/
void CMod_LoadLeafFaces (lump_t *l)
{
	int		i, j, count;
	int		*in;
	int		*out;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count > MAX_MAP_LEAFFACES)
		Com_Error (ERR_DROP, "Map has too many leaffaces");

	out = Z_Malloc (l->filelen);
	map_leaffaces = out;
	numleaffaces = count;

	for ( i=0 ; i<count ; i++)
	{
		j = LittleLong ( in[i] );

		if (j < 0 ||  j >= numfaces)
			Com_Error (ERR_DROP, "CMod_LoadLeafFaces: bad surface number");

		out[i] = j;
	}
}

/*
=================
CMod_LoadLeafs
=================
*/
void CMod_LoadLeafs (lump_t *l)
{
	int			i, j, k;
	cleaf_t		*out;
	dleaf_t 	*in;
	cbrush_t	*brush;
	cface_t		*face;
	int			count;
	int			numleaffaces;
	int			firstleafface;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map with no leafs");
	else if (count > MAX_MAP_LEAFS)
		Com_Error (ERR_DROP, "Map has too many leafs");

	out = map_leafs;
	numleafs = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->cluster = LittleLong ( in->cluster );
		out->area = LittleLong ( in->area ) + 1;
		out->firstleafbrush = LittleLong ( in->firstleafbrush );
		out->numleafbrushes = LittleLong ( in->numleafbrushes );

		out->contents = 0;
		out->firstleafpatch = numleafpatches;
		out->numleafpatches = 0;

		// OR brushes' contents
		for ( j = 0; j < out->numleafbrushes; j++ ) {
			brush = &map_brushes[map_leafbrushes[out->firstleafbrush + j]];
			out->contents |= brush->contents;
		}

		firstleafface = LittleLong ( in->firstleafface );
		numleaffaces = LittleLong ( in->numleaffaces );

		// OR patches' contents
		for ( j = 0; j < numleaffaces; j++ ) {
			k = map_leaffaces[firstleafface + j];
			face = &map_faces[k];

			if ( face->numbrushes ) {
				out->contents |= face->surface->contents;
				map_leafpatches[numleafpatches++] = k;
				out->numleafpatches++;
			}
		}

		if ( out->area >= numareas ) {
			map_areas[out->area].floodnum = 0;
			map_areas[out->area].floodvalid = 0;
			map_areas[out->area].numareaportals = 0;
			numareas = out->area + 1;
		}
	}
}

/*
=================
CMod_LoadPlanes
=================
*/
void CMod_LoadPlanes (lump_t *l)
{
	int			i, j;
	cplane_t	*out;
	dplane_t 	*in;
	int			count;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map with no planes");
	else if (count > MAX_MAP_PLANES)
		Com_Error (ERR_DROP, "Map has too many planes");

	out = map_planes;
	numplanes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->signbits = 0;
		out->type = PLANE_NONAXIAL;

		for (j=0 ; j<3 ; j++)
		{
			out->normal[j] = LittleFloat (in->normal[j]);
			if (out->normal[j] < 0)
				out->signbits |= 1<<j;
			if (out->normal[j] == 1.0f)
				out->type = j;
		}

		out->dist = LittleFloat (in->dist);

		CM_AddPlaneToHash ( CM_HashKeyForPlane (out->dist), out );
	}
}

/*
=================
CMod_LoadLeafBrushes
=================
*/
void CMod_LoadLeafBrushes (lump_t *l)
{
	int			i;
	int			*out;
	int		 	*in;
	int			count;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map with no planes");
	else if (count > MAX_MAP_LEAFBRUSHES)
		Com_Error (ERR_DROP, "Map has too many leafbrushes");

	out = map_leafbrushes;
	numleafbrushes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
		*out = LittleLong (*in);
}

/*
=================
CMod_LoadBrushSides
=================
*/
void CMod_LoadBrushSides (lump_t *l)
{
	int			i, j;
	cbrushside_t	*out;
	dbrushside_t 	*in;
	int			count;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count > MAX_MAP_BRUSHSIDES)
		Com_Error (ERR_DROP, "Map has too many brushsides");

	out = map_brushsides;
	numbrushsides = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->plane = map_planes + LittleLong (in->planenum);
		j = LittleLong (in->shadernum);
		if (j >= numshaderrefs)
			Com_Error (ERR_DROP, "Bad brushside texinfo");
		out->surface = &map_surfaces[j];
	}
}

/*
=================
CMod_LoadBrushes

=================
*/
void CMod_LoadBrushes (lump_t *l)
{
	dbrush_t	*in;
	cbrush_t	*out;
	int			i, count;
	int			shaderref;

	in = (void *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count > MAX_MAP_BRUSHES)
		Com_Error (ERR_DROP, "Map has too many brushes");

	out = map_brushes;
	numbrushes = count;

	for (i=0 ; i<count ; i++, out++, in++)
	{
		shaderref = LittleLong ( in->shadernum );
		out->contents = map_surfaces[shaderref].contents;
		out->firstbrushside = &map_brushsides[LittleLong (in->firstside)];
		out->numsides = LittleLong ( in->numsides );
	}
}

/*
=================
CMod_LoadVisibility
=================
*/
void CMod_LoadVisibility (lump_t *l)
{
	if (l->filelen > MAX_MAP_VISIBILITY)
		Com_Error (ERR_DROP, "Map has too large visibility lump");
	if (!l->filelen)
	{
		map_pvs->numclusters = 0;
		map_pvs->rowsize = MAX_CM_LEAFS / 8;
		return;
	}

	memcpy (map_visibility, cmod_base + l->fileofs, l->filelen);

	map_pvs->numclusters = LittleLong (map_pvs->numclusters);
	map_pvs->rowsize = LittleLong (map_pvs->rowsize);
}


/*
=================
CMod_LoadEntityString
=================
*/
void CMod_LoadEntityString (lump_t *l)
{
	numentitychars = l->filelen;
	if (l->filelen > MAX_MAP_ENTSTRING)
		Com_Error (ERR_DROP, "Map has too large entity lump");

	memcpy (map_entitystring, cmod_base + l->fileofs, l->filelen);
}

/*
=================
CM_CalcPHS
=================
*/
void CM_CalcPHS (void)
{
	int		rowbytes, rowwords;
	int		i, j, k, l, index;
	int		bitbyte;
	unsigned	*dest, *src;
	qbyte	*scan;
	int		count, vcount;
	int		numclusters;

	if ( !map_pvs->numclusters ) {
		return;
	}

	Com_DPrintf ("Building PHS...\n");

	rowwords = map_pvs->rowsize / sizeof(long);
	rowbytes = map_pvs->rowsize;

	memset ( map_phs, 0, MAX_CM_VISIBILITY );

	map_phs->rowsize = map_pvs->rowsize;
	map_phs->numclusters = numclusters = map_pvs->numclusters;

	vcount = 0;
	for (i=0 ; i<numclusters ; i++)
	{
		scan = CM_ClusterPVS ( i );
		for (j=0 ; j<numclusters ; j++)
		{
			if ( scan[j>>3] & (1<<(j&7)) )
			{
				vcount++;
			}
		}
	}

	count = 0;
	scan = (qbyte *)map_pvs->data;
	dest = (unsigned *)((qbyte *)map_phs + 8);

	for (i=0 ; i<numclusters ; i++, dest += rowwords, scan += rowbytes)
	{
		memcpy (dest, scan, rowbytes);
		for (j=0 ; j<rowbytes ; j++)
		{
			bitbyte = scan[j];
			if (!bitbyte)
				continue;
			for (k=0 ; k<8 ; k++)
			{
				if (! (bitbyte & (1<<k)) )
					continue;
				// OR this pvs row into the phs
				index = (j<<3) + k;
				if (index >= numclusters)
					Com_Error (ERR_DROP, "CM_CalcPHS: Bad bit in PVS");	// pad bits should be 0
				src = (unsigned *)((qbyte *)map_pvs->data) + index*rowwords;
				for (l=0 ; l<rowwords ; l++)
					dest[l] |= src[l];
			}
		}
		for (j=0 ; j<numclusters ; j++)
			if ( ((qbyte *)dest)[j>>3] & (1<<(j&7)) )
				count++;
	}

	Com_DPrintf ("Average clusters visible / hearable / total: %i / %i / %i\n"
		, vcount/numclusters, count/numclusters, numclusters);
}

/*
==================
CM_LoadMap

Loads in the map and all submodels
==================
*/
cmodel_t *CM_LoadMap (char *name, qboolean clientload, unsigned *checksum)
{
	unsigned		*buf;
	int				i;
	dheader_t		header;
	int				length;
	static unsigned	last_checksum;

	cm_noAreas = Cvar_Get ("cm_noAreas", "0", CVAR_CHEAT);
	cm_noCurves = Cvar_Get ("cm_noCurves", "0", CVAR_CHEAT);

	if ( !strcmp (map_name, name) && (clientload || !Cvar_VariableValue ("flushmap")) )
	{
		*checksum = last_checksum;
		if (!clientload)
		{
			memset (portalopen, 0, sizeof(portalopen));
			FloodAreaConnections ();
		}
		return &map_cmodels[0];		// still have the right version
	}

	// free old stuff
	numplanes = 0;
	numnodes = 0;
	numleafs = 0;
	numcmodels = 0;
	numentitychars = 0;
	numvertexes = 0;
	numfaces = 0;
	numleaffaces = 0;
	numleafpatches = 0;
	numareas = 1;
	numareaportals = 1;
	map_entitystring[0] = 0;
	map_name[0] = 0;

	if (!name || !name[0])
	{
		numleafs = 1;
		*checksum = 0;
		return &map_cmodels[0];			// cinematic servers won't have anything at all
	}

	//
	// load the file
	//
	length = FS_LoadFile (name, (void **)&buf);
	if (!buf)
		Com_Error (ERR_DROP, "Couldn't load %s", name);

	last_checksum = LittleLong (Com_BlockChecksum (buf, length));
	*checksum = last_checksum;

	header = *(dheader_t *)buf;
	for (i=0 ; i<sizeof(dheader_t)/4 ; i++)
		((int *)&header)[i] = LittleLong ( ((int *)&header)[i]);

	if ((header.version != Q3BSPVERSION) && (header.version != RTCWBSPVERSION))
		Com_Error (ERR_DROP, "CMod_LoadBrushModel: %s has wrong version number (%i should be %i or %i)"
		, name, header.version, Q3BSPVERSION, RTCWBSPVERSION);

	cmod_base = (qbyte *)buf;

	// load into heap
	CMod_LoadSurfaces (&header.lumps[LUMP_SHADERREFS]);
	CMod_LoadPlanes (&header.lumps[LUMP_PLANES]);
	CMod_LoadLeafBrushes (&header.lumps[LUMP_LEAFBRUSHES]);
	CMod_LoadBrushes (&header.lumps[LUMP_BRUSHES]);
	CMod_LoadBrushSides (&header.lumps[LUMP_BRUSHSIDES]);
	CMod_LoadVertexes (&header.lumps[LUMP_VERTEXES]);
	CMod_LoadFaces (&header.lumps[LUMP_FACES]);
	CMod_LoadLeafFaces (&header.lumps[LUMP_LEAFFACES]);
	CMod_LoadLeafs (&header.lumps[LUMP_LEAFS]);
	CMod_LoadNodes (&header.lumps[LUMP_NODES]);
	CMod_LoadSubmodels (&header.lumps[LUMP_MODELS]);
	CMod_LoadVisibility (&header.lumps[LUMP_VISIBILITY]);
	CMod_LoadEntityString (&header.lumps[LUMP_ENTITIES]);

	FS_FreeFile (buf);

	CM_InitBoxHull ();

	memset (portalopen, 0, sizeof(portalopen));
	FloodAreaConnections ();

	CM_CalcPHS ();

	CM_FreePlanesHash ();

	if (map_leaffaces) {
		Z_Free (map_leaffaces);
	}
	if (map_verts) {
		Z_Free (map_verts);
	}

	memset ( nullrow, 255, MAX_CM_LEAFS / 8 );

	Q_strncpyz ( map_name, name, sizeof(map_name) );

	return &map_cmodels[0];
}

/*
==================
CM_InlineModel
==================
*/
cmodel_t	*CM_InlineModel (char *name)
{
	int		num;

	if (!name || name[0] != '*')
		Com_Error (ERR_DROP, "CM_InlineModel: bad name");
	num = atoi (name+1);
	if (num < 1 || num >= numcmodels)
		Com_Error (ERR_DROP, "CM_InlineModel: bad number");

	return &map_cmodels[num];
}

int		CM_ClusterSize (void)
{
	return map_pvs->rowsize;
}

int		CM_NumClusters (void)
{
	return map_pvs->numclusters;
}

int		CM_NumInlineModels (void)
{
	return numcmodels;
}

char	*CM_EntityString (void)
{
	return map_entitystring;
}

int		CM_LeafContents (int leafnum)
{
	if (leafnum < 0 || leafnum >= numleafs)
		Com_Error (ERR_DROP, "CM_LeafContents: bad number");
	return map_leafs[leafnum].contents;
}

int		CM_LeafCluster (int leafnum)
{
	if (leafnum < 0 || leafnum >= numleafs)
		Com_Error (ERR_DROP, "CM_LeafCluster: bad number");
	return map_leafs[leafnum].cluster;
}

int		CM_LeafArea (int leafnum)
{
	if (leafnum < 0 || leafnum >= numleafs)
		Com_Error (ERR_DROP, "CM_LeafArea: bad number");
	return map_leafs[leafnum].area;
}

//=======================================================================

cplane_t	*box_planes;
cbrush_t	*box_brush;
cmodel_t	*box_cmodel;

/*
===================
CM_InitBoxHull

Set up the planes so that the six floats of a bounding box
can just be stored out and get a proper clipping hull structure.
===================
*/
void CM_InitBoxHull (void)
{
	int			i;
	int			side;
	cplane_t	*p;
	cbrushside_t	*s;

	if (numbrushes > MAX_CM_BRUSHES
		|| numbrushsides > MAX_CM_BRUSHSIDES
		|| numplanes > MAX_CM_PLANES
		|| numcmodels > MAX_CM_MODELS)
		Com_Error (ERR_DROP, "Not enough room for box tree");

	box_planes = &map_planes[numplanes];

	box_brush = &map_brushes[numbrushes];
	box_brush->numsides = 6;
	box_brush->firstbrushside = &map_brushsides[numbrushsides];
	box_brush->contents = CONTENTS_BODY;

	box_cmodel = &map_cmodels[numcmodels];
	box_cmodel->numfaces = 0;
	box_cmodel->firstface = 0;
	box_cmodel->firstbrush = numbrushes;
	box_cmodel->numbrushes = 1;

	for (i=0 ; i<6 ; i++)
	{
		side = i&1;

		// brush sides
		s = box_brush->firstbrushside + i;
		s->plane = box_planes + i;
		s->surface = &nullsurface;

		// planes
		p = &box_planes[i];
		VectorClear (p->normal);

		if ( i&1 ) {
			p->type = PLANE_NONAXIAL;
			p->normal[i>>1] = -1;
			p->signbits = (1<<(i>>1));
		} else {
			p->type = i>>1;
			p->normal[i>>1] = 1;
			p->signbits = 0;
		}
	}
}


/*
===================
CM_ModelForBBox

To keep everything totally uniform, bounding boxes are turned into inline models
===================
*/
cmodel_t *CM_ModelForBBox (vec3_t mins, vec3_t maxs)
{
	box_planes[0].dist = maxs[0];
	box_planes[1].dist = -mins[0];
	box_planes[2].dist = maxs[1];
	box_planes[3].dist = -mins[1];
	box_planes[4].dist = maxs[2];
	box_planes[5].dist = -mins[2];

	VectorCopy (mins, box_cmodel->mins);
	VectorCopy (maxs, box_cmodel->maxs);

	return box_cmodel;
}


/*
==================
CM_PointLeafnum_r

==================
*/
int CM_PointLeafnum_r (vec3_t p, int num)
{
	float		d;
	cnode_t		*node;
	cplane_t	*plane;

	while (num >= 0)
	{
		node = map_nodes + num;
		plane = node->plane;
		d = PlaneDiff (p, plane);

		if (d < 0)
			num = node->children[1];
		else
			num = node->children[0];
	}

	c_pointcontents++;		// optimize counter

	return -1 - num;
}

int CM_PointLeafnum (vec3_t p)
{
	if (!numplanes)
		return 0;		// sound may call this without map loaded
	return CM_PointLeafnum_r (p, 0);
}



/*
=============
CM_BoxLeafnums

Fills in a list of all the leafs touched
=============
*/
int		leaf_count, leaf_maxcount;
int		*leaf_list;
float	*leaf_mins, *leaf_maxs;
int		leaf_topnode;

void CM_BoxLeafnums_r (int nodenum)
{
	cnode_t	*node;
	int		s;

	while (1)
	{
		if (nodenum < 0)
		{
			if (leaf_count >= leaf_maxcount)
			{
				return;
			}
			leaf_list[leaf_count++] = -1 - nodenum;
			return;
		}

		node = &map_nodes[nodenum];
		s = BoxOnPlaneSide(leaf_mins, leaf_maxs, node->plane);

		if (s == 1)
			nodenum = node->children[0];
		else if (s == 2)
			nodenum = node->children[1];
		else
		{	// go down both
			if (leaf_topnode == -1)
				leaf_topnode = nodenum;
			CM_BoxLeafnums_r (node->children[0]);
			nodenum = node->children[1];
		}
	}
}

int	CM_BoxLeafnums (vec3_t mins, vec3_t maxs, int *list, int listsize, int *topnode)
{
	leaf_list = list;
	leaf_count = 0;
	leaf_maxcount = listsize;
	leaf_mins = mins;
	leaf_maxs = maxs;

	leaf_topnode = -1;

	CM_BoxLeafnums_r (0);

	if (topnode)
		*topnode = leaf_topnode;

	return leaf_count;
}


/*
==================
CM_BrushContents
==================
*/
static inline int CM_BrushContents ( cbrush_t *brush, vec3_t p )
{
	int				i;
	cbrushside_t	*brushside;

	brushside = brush->firstbrushside;
	for ( i = 0; i < brush->numsides; i++, brushside++ ) {
		if ( PlaneDiff (p, brushside->plane) > 0 )
			break;
	}

	if ( i == brush->numsides ) {
		return brush->contents;
	}

	return 0;
}

/*
==================
CM_FaceContents
==================
*/
static inline int CM_FaceContents ( cface_t *face, vec3_t p )
{
	int				i, c;
	cbrush_t		*brush;

	for ( i = 0, brush = face->brushes; i < face->numbrushes; i++ ) {
		if ( (c = CM_BrushContents (brush, p)) ) {
			return c;
		}
	}

	return 0;
}

/*
==================
CM_PointContents

==================
*/
int CM_PointContents ( vec3_t p, cmodel_t *cmodel )
{
	int				i, contents;
	cface_t			*face;
	cbrush_t		*brush;

	if ( !numnodes ) {	// map not loaded
		return 0;
	}

	if ( !cmodel || cmodel == map_cmodels ) {
		cleaf_t	*leaf = &map_leafs[CM_PointLeafnum_r (p, 0)];
		contents = ( leaf->contents & CONTENTS_NODROP );

		for ( i = 0; i < leaf->numleafbrushes; i++ ) {
			brush = &map_brushes[map_leafbrushes[leaf->firstleafbrush + i]];
			// check if brush actually adds something to contents
			if ( (contents & brush->contents) == brush->contents ) {
				continue;
			}
			contents |= CM_BrushContents ( brush, p );
		}

		if ( cm_noCurves->value || !leaf->numleafpatches ) {
			return contents;
		}

		for ( i = 0; i < leaf->numleafpatches; i++ ) {
			face = &map_faces[map_leafpatches[leaf->firstleafpatch + i]];
			// check if patch actually adds something to contents
			if ( (contents & face->surface->contents) == face->surface->contents ) {
				continue;
			}
			contents |= CM_FaceContents ( face, p );
		}
	} else {
		contents = 0;

		for ( i = 0; i < cmodel->numbrushes; i++ ) {
			brush = &map_brushes[cmodel->firstbrush + i];
			// check if brush actually adds something to contents
			if ( (contents & brush->contents) == brush->contents ) {
				continue;
			}
			contents |= CM_BrushContents ( brush, p );
		}

		if ( cm_noCurves->value || !cmodel->numfaces ) {
			return contents;
		}

		for ( i = 0; i < cmodel->numfaces; i++ ) {
			face = &map_faces[cmodel->firstface + i];
			// check if patch actually adds something to contents
			if ( !face->numbrushes || (contents & face->surface->contents) == face->surface->contents ) {
				continue;
			}
			contents |= CM_FaceContents ( face, p );
		}
	}

	return contents;
}

/*
==================
CM_TransformedPointContents

Handles offseting and rotation of the end points for moving and
rotating entities
==================
*/
int	CM_TransformedPointContents (vec3_t p, cmodel_t *cmodel, vec3_t origin, vec3_t angles)
{
	vec3_t		p_l;
	vec3_t		temp;
	mat3_t		axis;

	if (!numnodes)	// map not loaded
		return 0;

	// subtract origin offset
	VectorSubtract (p, origin, p_l);

	// rotate start and end into the models frame of reference
	if (cmodel && (cmodel != box_cmodel) && (angles[0] || angles[1] || angles[2]) )
	{
		AnglesToAxis (angles, axis);
		VectorCopy (p_l, temp);
		Matrix3_Multiply_Vec3 (axis, temp, p_l);
	}

	return CM_PointContents (p_l, cmodel);
}


/*
===============================================================================

BOX TRACING

===============================================================================
*/

// 1/32 epsilon to keep floating point happy
#define	DIST_EPSILON	(0.03125)

vec3_t	trace_start, trace_end;
vec3_t	trace_mins, trace_maxs;
vec3_t	trace_absmins, trace_absmaxs;
vec3_t	trace_extents;

trace_t	trace_trace;
int		trace_contents;
qboolean	trace_ispoint;		// optimized case


/*
================
CM_ClipBoxToBrush
================
*/
void CM_ClipBoxToBrush (vec3_t mins, vec3_t maxs, vec3_t p1, vec3_t p2,
					  trace_t *trace, cbrush_t *brush, float leave_epsilon)
{
	int			i, j;
	cplane_t	*plane, *clipplane;
	float		enterfrac, leavefrac;
	vec3_t		ofs;
	float		d1, d2;
	qboolean	getout, startout;
	float		f;
	cbrushside_t	*side, *leadside;

	enterfrac = -1;
	leavefrac = 1;
	clipplane = NULL;

	if (!brush->numsides)
		return;

	c_brush_traces++;

	getout = qfalse;
	startout = qfalse;
	leadside = NULL;
	side = brush->firstbrushside;

	for (i = 0; i < brush->numsides; i++, side++)
	{
		plane = side->plane;

		if (!trace_ispoint)
		{	// general box case

			// push the plane out apropriately for mins/maxs
			for (j=0 ; j<3 ; j++)
			{
				if (plane->normal[j] < 0)
					ofs[j] = maxs[j];
				else
					ofs[j] = mins[j];
			}

			if ( plane->type < 3 ) {
				d1 = p1[plane->type] + ofs[plane->type] - plane->dist;
				d2 = p2[plane->type] + ofs[plane->type] - plane->dist;
			} else {
				d1 = (p1[0] + ofs[0])*plane->normal[0] + (p1[1] + ofs[1])*plane->normal[1] + (p1[2] + ofs[2])*plane->normal[2] - plane->dist;
				d2 = (p2[0] + ofs[0])*plane->normal[0] + (p2[1] + ofs[1])*plane->normal[1] + (p2[2] + ofs[2])*plane->normal[2] - plane->dist;
			}
		}
		else
		{	// special point case
			if ( plane->type < 3 ) {
				d1 = p1[plane->type] - plane->dist;
				d2 = p2[plane->type] - plane->dist;
			} else {
				d1 = DotProduct (p1, plane->normal) - plane->dist;
				d2 = DotProduct (p2, plane->normal) - plane->dist;
			}
		}

		if (d2 > 0)
			getout = qtrue;	// endpoint is not in solid
		if (d1 > 0)
			startout = qtrue;

		// if completely in front of face, no intersection
		if (d1 > 0 && d2 >= d1)
			return;

		if (d1 <= 0 && d2 <= 0)
			continue;

		// crosses face
		if (d1 > d2)
		{	// enter
			f = (d1-DIST_EPSILON) / (d1-d2);
			if (f > enterfrac)
			{
				enterfrac = f;
				clipplane = plane;
				leadside = side;
			}
		}
		else
		{	// leave
			f = (d1+leave_epsilon) / (d1-d2);
			if (f < leavefrac)
				leavefrac = f;
		}
	}

	if (!startout)
	{	// original point was inside brush
		trace->startsolid = qtrue;
		if (!getout)
			trace->allsolid = qtrue;
		return;
	}

	if (enterfrac < leavefrac)
	{
		if (enterfrac > -1 && enterfrac < trace->fraction)
		{
			if (enterfrac < 0)
				enterfrac = 0;
			trace->fraction = enterfrac;
			trace->plane = *clipplane;
			trace->surface = leadside->surface;
			trace->contents = brush->contents;
		}
	}
}

/*
================
CM_TestBoxInBrush
================
*/
void CM_TestBoxInBrush (vec3_t mins, vec3_t maxs, vec3_t p1,
					  trace_t *trace, cbrush_t *brush)
{
	int			i, j;
	cplane_t	*plane;
	vec3_t		ofs;
	float		d1;
	cbrushside_t	*side;

	if (!brush->numsides)
		return;

	side = brush->firstbrushside;
	for (i = 0; i<brush->numsides; i++, side++)
	{
		plane = side->plane;

		// general box case

		// push the plane out apropriately for mins/maxs
		for (j=0 ; j<3 ; j++)
		{
			if (plane->normal[j] < 0)
				ofs[j] = maxs[j];
			else
				ofs[j] = mins[j];
		}

		if ( plane->type < 3 ) {
			d1 = p1[plane->type] + ofs[plane->type] - plane->dist;
		} else {
			d1 = (p1[0] + ofs[0])*plane->normal[0] + (p1[1] + ofs[1])*plane->normal[1] + (p1[2] + ofs[2])*plane->normal[2] - plane->dist;
		}

		// if completely in front of face, no intersection
		if (d1 > 0)
			return;
	}

	// inside this brush
	trace->startsolid = trace->allsolid = qtrue;
	trace->fraction = 0;
	trace->contents = brush->contents;
}

/*
================
CM_TraceToLeaf
================
*/
void CM_TraceToLeaf (int leafnum)
{
	int			i, j;
	cleaf_t		*leaf;
	cbrush_t	*b;
	cface_t		*face;

	leaf = &map_leafs[leafnum];
	if ( !(leaf->contents & trace_contents))
		return;

	// trace line against all brushes in the leaf
	for (i=0 ; i<leaf->numleafbrushes ; i++)
	{
		b = &map_brushes[map_leafbrushes[leaf->firstleafbrush+i]];
		if (b->checkcount == checkcount)
			continue;	// already checked this brush in another leaf
		b->checkcount = checkcount;
		if ( !(b->contents & trace_contents))
			continue;
		CM_ClipBoxToBrush (trace_mins, trace_maxs, trace_start, trace_end, &trace_trace, b, DIST_EPSILON);
		if (!trace_trace.fraction)
			return;
	}

	if (cm_noCurves->value || !leaf->numleafpatches)
		return;

	// trace line against all patches in the leaf
	for (i = 0; i < leaf->numleafpatches; i++)
	{
		face = &map_faces[map_leafpatches[leaf->firstleafpatch + i]];
		if (face->checkcount == checkcount)
			continue;	// already checked this patch
		face->checkcount = checkcount;
		if (!BoundsIntersect (face->mins, face->maxs, trace_absmins, trace_absmaxs))
			continue;
		if ( !(face->surface->contents & trace_contents) )
			continue;
		for (j = 0; j < face->numbrushes; j++)
		{
			CM_ClipBoxToBrush (trace_mins, trace_maxs, trace_start, trace_end, &trace_trace, &face->brushes[j], 0);
			if (!trace_trace.fraction)
				return;
		}
	}
}

/*
================
CM_TraceToInlineModel
================
*/
void CM_TraceToInlineModel (cmodel_t *mod)
{
	int			i, j;
	cbrush_t	*b;
	cface_t		*face;

	if (!BoundsIntersect (mod->mins, mod->maxs, trace_absmins, trace_absmaxs))
		return;

	// trace line against all brushes in the model
	for (i=0 ; i<mod->numbrushes ; i++)
	{
		b = &map_brushes[mod->firstbrush+i];
		if ( !(b->contents & trace_contents))
			continue;
		CM_ClipBoxToBrush (trace_mins, trace_maxs, trace_start, trace_end, &trace_trace, b, DIST_EPSILON);
		if (!trace_trace.fraction)
			return;
	}

	if (cm_noCurves->value || !mod->numfaces)
		return;

	// trace line against all patches in the model
	for (i = 0; i < mod->numfaces; i++)
	{
		face = &map_faces[mod->firstface+i];
		if (face->checkcount == checkcount)
			continue;	// already checked this patch
		face->checkcount = checkcount;
		if ( !face->numbrushes || !(face->surface->contents & trace_contents) )
			continue;
		for (j = 0; j < face->numbrushes; j++)
		{
			CM_ClipBoxToBrush (trace_mins, trace_maxs, trace_start, trace_end, &trace_trace, &face->brushes[j], 0);
			if (!trace_trace.fraction)
				return;
		}
	}
}

/*
================
CM_TestInLeaf
================
*/
void CM_TestInLeaf (int leafnum)
{
	int			i, j;
	cleaf_t		*leaf;
	cbrush_t	*b;
	cface_t		*face;

	leaf = &map_leafs[leafnum];
	if ( !(leaf->contents & trace_contents))
		return;

	// trace line against all brushes in the leaf
	for (i=0 ; i<leaf->numleafbrushes ; i++)
	{
		b = &map_brushes[map_leafbrushes[leaf->firstleafbrush+i]];
		if (b->checkcount == checkcount)
			continue;	// already checked this brush in another leaf
		b->checkcount = checkcount;
		if ( !(b->contents & trace_contents))
			continue;
		CM_TestBoxInBrush (trace_mins, trace_maxs, trace_start, &trace_trace, b);
		if (!trace_trace.fraction)
			return;
	}

	if (cm_noCurves->value || !leaf->numleafpatches)
		return;

	// trace line against all patches in the leaf
	for (i = 0; i < leaf->numleafpatches; i++)
	{
		face = &map_faces[map_leafpatches[leaf->firstleafpatch + i]];
		if (face->checkcount == checkcount)
			continue;	// already checked this patch
		face->checkcount = checkcount;
		if (!BoundsIntersect (face->mins, face->maxs, trace_absmins, trace_absmaxs))
			continue;
		if ( !(face->surface->contents & trace_contents) )
			continue;
		for (j = 0; j < face->numbrushes; j++)
		{
			CM_TestBoxInBrush (trace_mins, trace_maxs, trace_start, &trace_trace, &face->brushes[j]);
			if (!trace_trace.fraction)
				return;
		}
	}
}

/*
================
CM_TestInInlineModel
================
*/
void CM_TestInInlineModel (cmodel_t *mod)
{
	int			i, j;
	cbrush_t	*b;
	cface_t		*face;

	if (!BoundsIntersect (mod->mins, mod->maxs, trace_absmins, trace_absmaxs))
		return;

	// trace line against all brushes in the model
	for (i=0 ; i<mod->numbrushes ; i++)
	{
		b = &map_brushes[mod->firstbrush+i];
		if ( !(b->contents & trace_contents))
			continue;
		CM_TestBoxInBrush (trace_mins, trace_maxs, trace_start, &trace_trace, b);
		if (!trace_trace.fraction)
			return;
	}

	if (cm_noCurves->value || !mod->numfaces)
		return;

	// trace line against all patches in the model
	for (i = 0; i < mod->numfaces; i++)
	{
		face = &map_faces[mod->firstface+i];
		if (face->checkcount == checkcount)
			continue;	// already checked this patch
		face->checkcount = checkcount;
		if ( !face->numbrushes || !(face->surface->contents & trace_contents) )
			continue;
		for (j = 0; j < face->numbrushes; j++)
		{
			CM_TestBoxInBrush (trace_mins, trace_maxs, trace_start, &trace_trace, &face->brushes[j]);
			if (!trace_trace.fraction)
				return;
		}
	}
}

/*
==================
CM_RecursiveHullCheck

==================
*/
void CM_RecursiveHullCheck (int num, float p1f, float p2f, vec3_t p1, vec3_t p2)
{
	cnode_t		*node;
	cplane_t	*plane;
	float		t1, t2, offset;
	float		frac, frac2;
	float		idist;
	int			i;
	vec3_t		mid;
	int			side;
	float		midf;

	if (trace_trace.fraction <= p1f)
		return;		// already hit something nearer

	// if < 0, we are in a leaf node
	if (num < 0)
	{
		CM_TraceToLeaf (-1-num);
		return;
	}

	//
	// find the point distances to the seperating plane
	// and the offset for the size of the box
	//
	node = map_nodes + num;
	plane = node->plane;

	if (plane->type < 3)
	{
		t1 = p1[plane->type] - plane->dist;
		t2 = p2[plane->type] - plane->dist;
		offset = trace_extents[plane->type];
	}
	else
	{
		t1 = DotProduct (plane->normal, p1) - plane->dist;
		t2 = DotProduct (plane->normal, p2) - plane->dist;
		if (trace_ispoint)
			offset = 0;
		else
			offset = fabs(trace_extents[0]*plane->normal[0]) +
				fabs(trace_extents[1]*plane->normal[1]) +
				fabs(trace_extents[2]*plane->normal[2]);
	}


	// see which sides we need to consider
	if (t1 >= offset && t2 >= offset)
	{
		CM_RecursiveHullCheck (node->children[0], p1f, p2f, p1, p2);
		return;
	}
	if (t1 < -offset && t2 < -offset)
	{
		CM_RecursiveHullCheck (node->children[1], p1f, p2f, p1, p2);
		return;
	}

	// put the crosspoint DIST_EPSILON pixels on the near side
	if (t1 < t2)
	{
		idist = 1.0/(t1-t2);
		side = 1;
		frac2 = (t1 + offset + DIST_EPSILON)*idist;
		frac = (t1 - offset + DIST_EPSILON)*idist;
	}
	else if (t1 > t2)
	{
		idist = 1.0/(t1-t2);
		side = 0;
		frac2 = (t1 - offset - DIST_EPSILON)*idist;
		frac = (t1 + offset + DIST_EPSILON)*idist;
	}
	else
	{
		side = 0;
		frac = 1;
		frac2 = 0;
	}

	// move up to the node
	clamp ( frac, 0, 1 );

	midf = p1f + (p2f - p1f)*frac;
	for (i=0 ; i<3 ; i++)
		mid[i] = p1[i] + frac*(p2[i] - p1[i]);

	CM_RecursiveHullCheck (node->children[side], p1f, midf, p1, mid);

	// go past the node
	clamp ( frac2, 0, 1 );

	midf = p1f + (p2f - p1f)*frac2;
	for (i=0 ; i<3 ; i++)
		mid[i] = p1[i] + frac2*(p2[i] - p1[i]);

	CM_RecursiveHullCheck (node->children[side^1], midf, p2f, mid, p2);
}



//======================================================================

/*
==================
CM_BoxTrace
==================
*/
trace_t		CM_BoxTrace (vec3_t start, vec3_t end,
						  vec3_t mins, vec3_t maxs,
						  cmodel_t *cmodel, int brushmask)
{
	int		i;
	vec3_t	point;
	qboolean notworld = (cmodel && (cmodel != map_cmodels));

	checkcount++;		// for multi-check avoidance

	c_traces++;			// for statistics, may be zeroed

	// fill in a default trace
	memset (&trace_trace, 0, sizeof(trace_trace));
	trace_trace.fraction = 1;
	trace_trace.surface = &nullsurface;

	if (!numnodes)	// map not loaded
		return trace_trace;

	trace_contents = brushmask;
	VectorCopy (start, trace_start);
	VectorCopy (end, trace_end);
	VectorCopy (mins, trace_mins);
	VectorCopy (maxs, trace_maxs);

	// build a bounding box of the entire move
	ClearBounds (trace_absmins, trace_absmaxs);
	VectorAdd (start, trace_mins, point);
	AddPointToBounds (point, trace_absmins, trace_absmaxs);
	VectorAdd (start, trace_maxs, point);
	AddPointToBounds (point, trace_absmins, trace_absmaxs);
	VectorAdd (end, trace_mins, point);
	AddPointToBounds (point, trace_absmins, trace_absmaxs);
	VectorAdd (end, trace_maxs, point);
	AddPointToBounds (point, trace_absmins, trace_absmaxs);

	//
	// check for position test special case
	//
	if (start[0] == end[0] && start[1] == end[1] && start[2] == end[2])
	{
		int		leafs[1024];
		int		i, numleafs;
		vec3_t	c1, c2;
		int		topnode;

		if (notworld)
		{
			CM_TestInInlineModel (cmodel);
		}
		else
		{
			VectorAdd (start, mins, c1);
			VectorAdd (start, maxs, c2);
			for (i=0 ; i<3 ; i++)
			{
				c1[i] -= 1;
				c2[i] += 1;
			}

			numleafs = CM_BoxLeafnums (c1, c2, leafs, 1024, &topnode);
			for (i=0 ; i<numleafs ; i++)
			{
				CM_TestInLeaf (leafs[i]);
				if (trace_trace.allsolid)
					break;
			}
		}

		VectorCopy (start, trace_trace.endpos);
		return trace_trace;
	}

	//
	// check for point special case
	//
	if (mins[0] == 0 && mins[1] == 0 && mins[2] == 0
		&& maxs[0] == 0 && maxs[1] == 0 && maxs[2] == 0)
	{
		trace_ispoint = qtrue;
		VectorClear (trace_extents);
	}
	else
	{
		trace_ispoint = qfalse;
		trace_extents[0] = -mins[0] > maxs[0] ? -mins[0] : maxs[0];
		trace_extents[1] = -mins[1] > maxs[1] ? -mins[1] : maxs[1];
		trace_extents[2] = -mins[2] > maxs[2] ? -mins[2] : maxs[2];
	}

	//
	// general sweeping through world
	//
	if (notworld)
		CM_TraceToInlineModel (cmodel);
	else
		CM_RecursiveHullCheck (0, 0, 1, start, end);

	if (trace_trace.fraction == 1)
	{
		VectorCopy (end, trace_trace.endpos);
	}
	else
	{
		for (i=0 ; i<3 ; i++)
			trace_trace.endpos[i] = start[i] + trace_trace.fraction * (end[i] - start[i]);
	}
	return trace_trace;
}


/*
==================
CM_TransformedBoxTrace

Handles offseting and rotation of the end points for moving and
rotating entities
==================
*/
trace_t		CM_TransformedBoxTrace (vec3_t start, vec3_t end,
						  vec3_t mins, vec3_t maxs,
						  cmodel_t *cmodel, int brushmask,
						  vec3_t origin, vec3_t angles)
{
	trace_t		trace;
	vec3_t		start_l, end_l;
	vec3_t		a, temp;
	mat3_t		axis;
	qboolean	rotated;

	// subtract origin offset
	VectorSubtract (start, origin, start_l);
	VectorSubtract (end, origin, end_l);

	// rotate start and end into the models frame of reference
	if (cmodel && (cmodel != box_cmodel) && (angles[0] || angles[1] || angles[2]) )
		rotated = qtrue;
	else
		rotated = qfalse;

	if (rotated)
	{
		AnglesToAxis (angles, axis);

		VectorCopy (start_l, temp);
		Matrix3_Multiply_Vec3 (axis, temp, start_l);

		VectorCopy (end_l, temp);
		Matrix3_Multiply_Vec3 (axis, temp, end_l);
	}

	// sweep the box through the model
	trace = CM_BoxTrace (start_l, end_l, mins, maxs, cmodel, brushmask);

	if (rotated && trace.fraction != 1.0)
	{
		VectorNegate (angles, a);
		AnglesToAxis (a, axis);

		VectorCopy (trace.plane.normal, temp);
		Matrix3_Multiply_Vec3 (axis, temp, trace.plane.normal);
	}

	if (trace_trace.fraction == 1)
	{
		VectorCopy (end, trace.endpos);
	}
	else
	{
		trace.endpos[0] = start[0] + trace.fraction * (end[0] - start[0]);
		trace.endpos[1] = start[1] + trace.fraction * (end[1] - start[1]);
		trace.endpos[2] = start[2] + trace.fraction * (end[2] - start[2]);
	}

	return trace;
}


/*
===============================================================================

PVS / PHS

===============================================================================
*/

qbyte	*CM_ClusterPVS (int cluster)
{
	if (cluster != -1 && map_pvs->numclusters)
		return (qbyte *)map_pvs->data + cluster * map_pvs->rowsize;

	return nullrow;
}

qbyte	*CM_ClusterPHS (int cluster)
{
	if (cluster != -1 && map_phs->numclusters)
		return (qbyte *)map_phs->data + cluster * map_phs->rowsize;

	return nullrow;
}

/*
===============================================================================

AREAPORTALS

===============================================================================
*/

void CM_AddAreaPortal ( int portalnum, int area, int otherarea )
{
	int i;
	carea_t *a;
	careaportal_t *ap;

	if ( portalnum >= MAX_CM_PORTALS ) {
		return;
	}
	if ( !area || area > numareas || !otherarea || otherarea > numareas ) {
		return;
	}

	ap = map_areaportals;
	for ( i = 0; i < numareaportals; i++, ap++ ) {
		if ( ap->portalnum == portalnum ) {
			return;		// already exists
		}
	}

	if ( numareaportals+2 > MAX_CM_AREAPORTALS ) {
		return;
	}

	ap = &map_areaportals[numareaportals];
	ap->portalnum = portalnum;
	ap->otherarea = otherarea;

	a = &map_areas[area];
	a->areaportals[a->numareaportals++] = numareaportals++;

	ap = &map_areaportals[numareaportals];
	ap->portalnum = portalnum;
	ap->otherarea = area;

	a = &map_areas[otherarea];
	a->areaportals[a->numareaportals++] = numareaportals++;
}

void FloodArea_r (carea_t *area, int floodnum)
{
	int		i;
	careaportal_t	*p;

	if (area->floodvalid == floodvalid)
	{
		if (area->floodnum == floodnum)
			return;
		Com_Error (ERR_DROP, "FloodArea_r: reflooded");
	}

	area->floodnum = floodnum;
	area->floodvalid = floodvalid;
	for (i=0 ; i<area->numareaportals ; i++)
	{
		p = &map_areaportals[area->areaportals[i]];
		if (portalopen[p->portalnum])
			FloodArea_r (&map_areas[p->otherarea], floodnum);
	}
}

/*
====================
FloodAreaConnections


====================
*/
void	FloodAreaConnections (void)
{
	int		i;
	carea_t	*area;
	int		floodnum;

	// all current floods are now invalid
	floodvalid++;
	floodnum = 0;

	// area 0 is not used
	for (i=1 ; i<numareas ; i++)
	{
		area = &map_areas[i];
		if (area->floodvalid == floodvalid)
			continue;		// already flooded into
		floodnum++;
		FloodArea_r (area, floodnum);
	}

}

void	CM_SetAreaPortalState (int portalnum, qboolean open)
{
	if (portalnum > MAX_CM_PORTALS)
		Com_Error (ERR_DROP, "areaportal > numareaportals");

	portalopen[portalnum] = open;
	FloodAreaConnections ();
}

qboolean	CM_AreasConnected (int area1, int area2)
{
	if (cm_noAreas->value)
		return qtrue;

	if (area1 > numareas || area2 > numareas)
		Com_Error (ERR_DROP, "CM_AreasConnected: area > numareas");

	if (map_areas[area1].floodnum == map_areas[area2].floodnum)
		return qtrue;
	return qfalse;
}


/*
=================
CM_WriteAreaBits

Writes a length byte followed by a bit vector of all the areas
that area in the same flood as the area parameter

This is used by the client refreshes to cull visibility
=================
*/
int CM_WriteAreaBits (qbyte *buffer, int area)
{
	int		i;
	int		bytes;

	bytes = (numareas+7)>>3;

	if (cm_noAreas->value)
	{	// for debugging, send everything
		memset (buffer, 255, bytes);
	}
	else
	{
		memset (buffer, 0, bytes);

		for (i=1 ; i<numareas ; i++)
		{
			if (!area || CM_AreasConnected ( i, area ) || i == area)
				buffer[i>>3] |= 1<<(i&7);
		}
	}

	return bytes;
}

/*
=================
CM_MergeAreaBits
=================
*/
void CM_MergeAreaBits (qbyte *buffer, int area)
{
	int		i;

	for (i=1 ; i<numareas ; i++)
	{
		if ( CM_AreasConnected ( i, area ) || i == area)
			buffer[i>>3] |= 1<<(i&7);
	}
}

/*
===================
CM_WritePortalState

Writes the portal state to a savegame file
===================
*/
void	CM_WritePortalState (FILE *f)
{
	int i, j;

	fwrite ( portalopen, sizeof(portalopen), 1, f );

	fwrite ( &numareaportals, sizeof(int), 1, f );
	fwrite ( map_areaportals, sizeof(map_areaportals[0])*numareaportals, 1, f );

	fwrite ( &numareas, sizeof(int), 1, f );

	for ( i = 1; i < numareas; i++ ) {
		fwrite ( &map_areas[i].numareaportals, sizeof(int), 1, f );

		for ( j = 0; j < map_areas[i].numareaportals; j++ ) {
			fwrite ( &map_areas[i].areaportals[j], sizeof(int), 1, f );
		}
	}
}

/*
===================
CM_ReadPortalState

Reads the portal state from a savegame file
and recalculates the area connections
===================
*/
void	CM_ReadPortalState (FILE *f)
{
	int i, j;

	fread ( portalopen, sizeof(portalopen), 1, f );

	fread ( &numareaportals, sizeof(int), 1, f );
	fread ( map_areaportals, sizeof(map_areaportals[0])*numareaportals, 1, f );

	fread ( &numareas, sizeof(int), 1, f );

	for ( i = 1; i < numareas; i++ ) {
		fread ( &map_areas[i].numareaportals, sizeof(int), 1, f );

		for ( j = 0; j < map_areas[i].numareaportals; j++ ) {
			fread ( &map_areas[i].areaportals[j], sizeof(int), 1, f );
		}
	}

	FloodAreaConnections ();
}

/*
=============
CM_HeadnodeVisible

Returns true if any leaf under headnode has a cluster that
is potentially visible
=============
*/
qboolean CM_HeadnodeVisible (int nodenum, qbyte *visbits)
{
	int		leafnum;
	int		cluster;
	cnode_t	*node;

	if (nodenum < 0)
	{
		leafnum = -1-nodenum;
		cluster = map_leafs[leafnum].cluster;
		if (cluster == -1)
			return qfalse;
		if (visbits[cluster>>3] & (1<<(cluster&7)))
			return qtrue;
		return qfalse;
	}

	node = &map_nodes[nodenum];
	if (CM_HeadnodeVisible(node->children[0], visbits))
		return qtrue;
	return CM_HeadnodeVisible(node->children[1], visbits);
}

